from __future__ import print_function
import numpy as np
from scipy.optimize import minimize as scipimin
from collections import OrderedDict

def compute_saxs(q,params):
    """
    Given q and a dict of parameters,
    compute the saxs spectrum.
    Supported parameters are the same as 
    SaxsParameterization Operation outputs,
    and should include at least the following keys:
    I_at_0, precursor_flag, form_flag, structure_flag.

    TODO: Document the equation.
    """
    pre_flag = params['precursor_flag']
    f_flag = params['form_flag']
    s_flag = params['structure_flag']
    I_at_0 = params['I_at_0']
    if not any([pre_flag,f_flag]):
        return np.ones(len(q))*I_at_0
    else:
        I = np.zeros(len(q))
        if pre_flag:
            I0_pre = params['I0_pre']
            r0_pre = params['r0_pre']
            I_pre = compute_spherical_normal_saxs(q,r0_pre,0)
            I += I0_pre*I_pre
        if f_flag: 
            I0_sph = params['I0_form']
            r0_sph = params['r0_form']
            sigma_sph = params['sigma_form']
            I_sph = compute_spherical_normal_saxs(q,r0_sph,sigma_sph)
            I += I0_sph*I_sph
        #if s_flag:
        #    I0_pk = params['I0_pk']
        #    I_pk = compute_peaks()
    return I

def compute_spherical_normal_saxs(q,r0,sigma):
    """
    Given q, a mean radius r0, 
    and the fractional standard deviation of radius sigma,
    compute the saxs spectrum assuming spherical particles 
    with normal size distribution.
    The returned intensity is normalized 
    such that I(q=0) is equal to 1.
    """
    q_zero = (q == 0)
    q_nz = np.invert(q_zero) 
    I = np.zeros(q.shape)
    if sigma < 1E-9:
        x = q*r0
        V_r0 = float(4)/3*np.pi*r0**3
        I[q_nz] = V_r0**2 * (3.*(np.sin(x[q_nz])-x[q_nz]*np.cos(x[q_nz]))*x[q_nz]**-3)**2
        I_zero = V_r0**2 
    else:
        sigma_r = sigma*r0
        dr = sigma_r*0.02
        rmin = np.max([r0-5*sigma_r,dr])
        rmax = r0+5*sigma_r
        I_zero = 0
        for ri in np.arange(rmin,rmax,dr):
            xi = q*ri
            V_ri = float(4)/3*np.pi*ri**3
            # The normal-distributed density of particles with radius r_i:
            rhoi = 1./(np.sqrt(2*np.pi)*sigma_r)*np.exp(-1*(r0-ri)**2/(2*sigma_r**2))
            I_zero += V_ri**2 * rhoi*dr
            I[q_nz] += V_ri**2 * rhoi*dr*(3.*(np.sin(xi[q_nz])-xi[q_nz]*np.cos(xi[q_nz]))*xi[q_nz]**-3)**2
    if any(q_zero):
        I[q_zero] = I_zero
    I = I/I_zero 
    return I

def precursor_heuristics(q,I,I_at_0=None):
    """
    Makes an educated guess for the radius of a small scatterer
    that would produce the input q, I(q).
    Result is bounded between 0 and 10 Angstroms.
    """
    n_q = len(q)
    # optimize the pearson correlation in the upper half of the q domain
    fit_obj = lambda r: -1*compute_pearson(compute_spherical_normal_saxs(q[n_q/2:],r,0),I[n_q/2:])
    #res = scipimin(fit_ojb,[0.1],bounds=[(0,0.3)]) 
    res = scipimin(fit_obj,[5],bounds=[(0,10)])
    r_opt = res.x[0]
    return r_opt
    # Assume the first dip of the precursor form factor occurs around the max of our q range.
    # First dip = qmax*r_pre ~ 4.5
    #r0_pre = 4.5/q[-1] 

def spherical_normal_heuristics(q,I,I_at_0=None):
    """
    This algorithm was developed and 
    originally contributed by Amanda Fournier.    

    Performs some heuristic measurements on the input spectrum,
    in order to make educated guesses 
    for the parameters of a size distribution
    (mean and standard deviation of radius)
    for a population of spherical scatterers.

    TODO: Document algorithm here.
    """
    if I_at_0 is None:
        I_at_0 = fit_I0(q,I)
    m = saxs_Iq4_metrics(q,I)
    width_metric = m['pI_qwidth']/m['q_at_Iqqqq_min1']
    intensity_metric = m['I_at_Iqqqq_min1']/I_at_0
    #######
    #
    # POLYNOMIALS FITTED FOR q0+/-10%,
    # where q0 is the argmin of a parabola
    # that is fit around the first minimum of I*q**4.
    # The function spherical_normal_heuristics_setup()
    # (in this same module) can be used to generate these polynomials.
    # polynomial for qr0 focus (with x=sigma_r/r0):
    # -8.05459639763x^2 + -0.470989868709x + 4.50108683096
    p_qr0_focus = [-8.05459639763,-0.470989868709,4.50108683096]
    # polynomial for width metric (with x=sigma_r/r0):
    # 3.12889797288x^2 + -0.0645231661487x + 0.0576604958693
    p_w = [3.12889797288,-0.0645231661487,0.0576604958693]
    # polynomial for intensity metric (with x=sigma_r/r0):
    # -1.33327411025x^3 + 0.432533640102x^2 + 0.00263776123775x + -1.27646761062e-05
    p_I = [-1.33327411025,0.432533640102,0.00263776123775,-1.27646761062e-05]
    #
    #######
    # Now find the sigma_r/r0 value that gets the extracted metrics
    # as close as possible to p_I and p_w.
    width_error = lambda x: (np.polyval(p_w,x)-width_metric)**2
    intensity_error = lambda x: (np.polyval(p_I,x)-intensity_metric)**2
    # TODO: make the objective function weight all errors equally
    heuristics_error = lambda x: width_error(x) + intensity_error(x)
    res = scipimin(heuristics_error,[0.1],bounds=[(0,0.3)]) 
    if not res.success:
        self.outputs['return_code'] = 2 
        msg = str('[{}] function minimization failed during '
        + 'form factor parameter extraction.'.format(__name__))
        self.outputs['features']['message'] = msg
    sigma_over_r = res.x[0]
    qr0_focus = np.polyval(p_qr0_focus,sigma_over_r)
    # qr0_focus = x1  ==>  r0 = x1 / q1
    r0 = qr0_focus/m['q_at_Iqqqq_min1']
    #sigma_r = sigma_over_r * r0 
    return r0,sigma_over_r

def saxs_Iq4_metrics(q,I):
    """
    From an input spectrum q and I(q),
    compute several properties of the I(q)*q^4 curve.
    This was designed for spectra that are 
    dominated by a dilute spherical form factor term.
    The metrics extracted by this Operation
    were originally intended as an intermediate step
    for estimating size distribution parameters 
    for a population of dilute spherical scatterers.

    Returns a dict of metrics.
    Dict keys and meanings:
    q_at_Iqqqq_min1: q value at first minimum of I*q^4
    I_at_Iqqqq_min1: I value at first minimum of I*q^4
    Iqqqq_min1: I*q^4 value at first minimum of I*q^4
    pIqqqq_qwidth: Focal q-width of polynomial fit to I*q^4 near first minimum of I*q^4 
    pIqqqq_Iqqqqfocus: Focal point of polynomial fit to I*q^4 near first minimum of I*q^4
    pI_qvertex: q value of vertex of polynomial fit to I(q) near first minimum of I*q^4  
    pI_Ivertex: I(q) at vertex of polynomial fit to I(q) near first minimum of I*q^4
    pI_qwidth: Focal q-width of polynomial fit to I(q) near first minimum of I*q^4
    pI_Iforcus: Focal point of polynomial fit to I(q) near first minimum of I*q^4

    TODO: document the algorithm here.
    """
    d = {}
    #if not dI:
    #    # uniform weights
    #    wt = np.ones(q.shape)   
    #else:
    #    # inverse error weights, 1/dI, 
    #    # appropriate if dI represents
    #    # Gaussian uncertainty with sigma=dI
    #    wt = 1./dI
    #######
    # Heuristics step 1: Find the first local max
    # and subsequent local minimum of I*q**4 
    Iqqqq = I*q**4
    # w is the number of adjacent points to consider 
    # when examining the I*q^4 curve for local extrema.
    # A greater value of w filters out smaller extrema.
    w = 10
    idxmax1, idxmin1 = 0,0
    stop_idx = len(q)-w-1
    test_range = iter(range(w,stop_idx))
    idx = test_range.next() 
    while any([idxmax1==0,idxmin1==0]) and idx < stop_idx-1:
        if np.argmax(Iqqqq[idx-w:idx+w+1]) == w and idxmax1 == 0:
            idxmax1 = idx
        if np.argmin(Iqqqq[idx-w:idx+w+1]) == w and idxmin1 == 0 and not idxmax1 == 0:
            idxmin1 = idx
        idx = test_range.next()
    if idxmin1 == 0 or idxmax1 == 0:
        ex_msg = str('unable to find first maximum and minimum of I*q^4 '
        + 'by scanning for local extrema with a window width of {} points'.format(w))
        d['message'] = ex_msg 
        raise Exception(ex_msg)
    #######
    # Heuristics 2: Characterize I*q**4 around idxmin1, 
    # by locally fitting a standardized polynomial.
    idx_around_min1 = (q>0.9*q[idxmin1]) & (q<1.1*q[idxmin1])
    q_min1_mean = np.mean(q[idx_around_min1])
    q_min1_std = np.std(q[idx_around_min1])
    q_min1_s = (q[idx_around_min1]-q_min1_mean)/q_min1_std
    Iqqqq_min1_mean = np.mean(Iqqqq[idx_around_min1])
    Iqqqq_min1_std = np.std(Iqqqq[idx_around_min1])
    Iqqqq_min1_s = (Iqqqq[idx_around_min1]-Iqqqq_min1_mean)/Iqqqq_min1_std
    p_min1 = np.polyfit(q_min1_s,Iqqqq_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_at_min1 = -1*p_min1[1]/(2*p_min1[0])
    d['q_at_Iqqqq_min1'] = qs_at_min1*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Iqqqqs_at_min1 = np.polyval(p_min1,qs_at_min1)
    d['Iqqqq_min1'] = Iqqqqs_at_min1*Iqqqq_min1_std+Iqqqq_min1_mean
    d['I_at_Iqqqq_min1'] = d['Iqqqq_min1']*float(1)/(d['q_at_Iqqqq_min1']**4)
    # The focal width of the parabola is 1/a 
    p_min1_fwidth = float(1)/p_min1[0] 
    d['pIqqqq_qwidth'] = p_min1_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    p_min1_fpoint = Iqqqqs_at_min1+float(1)/(4*p_min1[0])
    d['pIqqqq_Iqqqqfocus'] = p_min1_fpoint*Iqqqq_min1_std+Iqqqq_min1_mean
    #######
    # Heuristics 2b: Characterize I(q) near min1 of I*q^4.
    I_min1_mean = np.mean(I[idx_around_min1])
    I_min1_std = np.std(I[idx_around_min1])
    I_min1_s = (I[idx_around_min1]-I_min1_mean)/I_min1_std
    pI_min1 = np.polyfit(q_min1_s,I_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_vertex = -1*pI_min1[1]/(2*pI_min1[0])
    d['pI_qvertex'] = qs_vertex*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Is_vertex = np.polyval(pI_min1,qs_vertex)
    d['pI_Ivertex'] = Is_vertex*I_min1_std+I_min1_mean
    # The focal width of the parabola is 1/a 
    pI_fwidth = float(1)/pI_min1[0]
    d['pI_qwidth'] = pI_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    pI_fpoint = Is_vertex+float(1)/(4*pI_min1[0])
    d['pI_Ifocus'] = pI_fpoint*I_min1_std+I_min1_mean
    #######
    return d

def compute_saxs_with_substitutions(q,d,x_keys,x_vals):
    d_sub = d.copy()
    for k,v in zip(x_keys,x_vals):
        d_sub[k] = v
    return compute_saxs(q,d_sub)


def saxs_fit(q,I,objfun,features,x_keys,constraints=[]):
    """
    Fit a saxs spectrum (I(q) vs q) to the theoretical spectrum 
    for one or several scattering populations.
    Input objfun (string) specifies objective function to use in optimization.
    Input features (dict) describes spectrum and scatterer populations.
    Input x_keys (list of strings) are the keys of variables that will be optimized.
    Every item in x_keys should be a key in the features dict.
    Input constraints (list of strings) to specify constraints.
    
    Supported objective functions: 
    (1) 'full_spectrum_chi2': sum of difference squared across entire q range. 
    (2) 'full_spectrum_chi2log': sum of difference of logarithm, squared, across entire q range. 
    (3) 'full_spectrum_chi2norm': sum of difference divided by measured value, squared, aross entire q range. 
    (4) 'low_q_chi2': sum of difference squared in only the lowest half of measured q range. 
    (5) 'low_q_chi2log': sum of difference of logarithm, squared, in lowest half of measured q range. 
    (6) 'pearson': pearson correlation between measured and modeled spectra. 
    (7) 'pearson_log': pearson correlation between logarithms of measured and modeled spectra.
    (8) 'low_q_pearson': pearson correlation between measured and modeled spectra. 
    (9) 'low_q_pearson_log': pearson correlation between logarithms of measured and modeled spectra. 

    Supported constraints: 
    (1) 'fix_I0': keeps I(q=0) fixed while fitting x_keys.

    TODO: document the objective functions, etc.
    """
    pre_flag = features['precursor_flag']
    form_flag = features['form_flag']
    structure_flag = features['structure_flag']

    # trim non-flagged populations out of x_keys
    if not pre_flag:
        if 'I0_pre' in x_keys:
            x_keys.pop(x_keys.index('I0_pre')) 
        if 'r0_pre' in x_keys:
            x_keys.pop(x_keys.index('r0_pre')) 
    if not form_flag:
        if 'I0_form' in x_keys:
            x_keys.pop(x_keys.index('I0_form')) 
        if 'r0_form' in x_keys:
            x_keys.pop(x_keys.index('r0_form')) 
        if 'sigma_form' in x_keys:
            x_keys.pop(x_keys.index('sigma_form')) 
    if not structure_flag:
        if 'I0_structure' in x_keys:
            x_keys.pop(x_keys.index('I0_structure')) 
        if 'q0_structure' in x_keys:
            x_keys.pop(x_keys.index('q0_structure')) 
        if 'sigma_structure' in x_keys:
            x_keys.pop(x_keys.index('sigma_structure')) 

    c = []
    if 'fix_I0' in constraints: 
        I_keys = []
        if 'I0_pre' in x_keys:
            I_keys.append('I0_pre')
        if 'I0_form' in x_keys:
            I_keys.append('I0_form')
        if 'I0_structure' in x_keys:
            I_keys.append('I0_structure')
        if len(I_keys) == 0:
            # constraint inherently satisfied: do nothing.
            pass
        elif len(I_keys) == 1:
            # overconstrained: skip the constraint, reduce dimensionality. 
            x_keys.pop(x_keys.index(I_keys[0])) 
        else:
            # find the indices of the relevant x_keys and form a constraint function.
            iargs = []
            Icons = 0
            for ik,k in zip(range(len(x_keys)),x_keys):
                if k in I_keys:
                    iargs.append(ik)
                    Icons += features[k]
            cfun = lambda x: sum([x[i] for i in iargs]) - Icons 
            c_fixI0 = {'type':'eq','fun':cfun}
            c.append(c_fixI0)

    x_init = [] 
    x_bounds = [] 
    for k in x_keys:
        #if k in features.keys():
        x_init.append(features[k])
        if k in ['r0_pre','r0_form','I0_pre','I0_form']:
            x_bounds.append((0,None))
        elif k in ['sigma_form']:
            x_bounds.append((0.0,0.3))

    saxs_fun = lambda q,x,d: compute_saxs_with_substitutions(q,d,x_keys,x)
    I_nz = (I>0)
    n_q = len(q)
    if objfun == 'full_spectrum_chi2':
        fit_obj = lambda x: np.sum( (saxs_fun(q,x,features) - I)**2 )
    elif objfun == 'full_spectrum_chi2log':
        fit_obj = lambda x: np.sum( (np.log(saxs_fun(q[I_nz],x,features)) - np.log(I[I_nz]))**2 )
    elif objfun == 'full_spectrum_chi2norm':
        fit_obj = lambda x: np.sum( ((saxs_fun(q[I_nz],x,features) - I[I_nz])/I[I_nz])**2 )
    elif objfun == 'low_q_chi2':
        fit_obj = lambda x: np.sum( (saxs_fun(q[:n_q/2],x,features) - I[:n_q/2])**2 )
    elif objfun == 'pearson':
        fit_obj = lambda x: -1*compute_pearson(saxs_fun(q,x,features), I)
    elif objfun == 'low_q_pearson':
        fit_obj = lambda x: -1*compute_pearson(saxs_fun(q[:n_q/2],x,features), I[:n_q/2]) 
    elif objfun == 'low_q_chi2log':
        fit_obj = lambda x: np.sum( (np.log(saxs_fun(q[I_nz][:n_q/2],x,features)) - np.log(I[I_nz][:n_q/2]))**2 )
    elif objfun == 'pearson_log':
        fit_obj = lambda x: -1*compute_pearson(np.log(saxs_fun(q[I_nz],x,features)), np.log(I[I_nz])) 
    elif objfun == 'low_q_pearson_log':
        fit_obj = lambda x: -1*compute_pearson(np.log(saxs_fun(q[I_nz][:n_q/2],x,features)), np.log(I[I_nz][:n_q/2])) 
    else:
        msg = 'objective function {} not supported'.format(objfun)
        raise ValueError(msg)
    res = scipimin(fit_obj,x_init,bounds=x_bounds,constraints=c)

    x_opt = res.x
    d_opt = OrderedDict()
    for k,xk in zip(x_keys,x_opt):
        d_opt[k] = xk
    d_opt['objective_before'] = fit_obj(x_init)
    d_opt['objective_after'] = fit_obj(x_opt)
    return d_opt    

def compute_pearson(y1,y2):
    y1mean = np.mean(y1)
    y2mean = np.mean(y2)
    y1std = np.std(y1)
    y2std = np.std(y2)
    return np.sum((y1-y1mean)*(y2-y2mean))/(np.sqrt(np.sum((y1-y1mean)**2))*np.sqrt(np.sum((y2-y2mean)**2)))

def fit_I0(q,I):
    """
    Find an estimate for I(q=0) by polynomial fitting.
    All of the input q, I(q) values are used in the fitting.
    """
    I_mean = np.mean(I)
    I_std = np.std(I)
    q_mean = np.mean(q)
    q_std = np.std(q)
    I_s = (I-I_mean)/I_std
    q_s = (q-q_mean)/q_std
    p = fit_with_slope_constraint(q_s,I_s,-1*q_mean/q_std,0,4) 
    I_at_0 = np.polyval(p,-1*q_mean/q_std)*I_std+I_mean
    return I_at_0

def fit_with_slope_constraint(q,I,q_cons,dIdq_cons,order,weights=None):
    """
    Perform a polynomial fitting 
    of the low-q region of the spectrum
    with dI/dq(q=0) constrained to be zero.
    This is performed by forming a Lagrangian 
    from a quadratic cost function 
    and the Lagrange-multiplied constraint function.
    
    TODO: Explicitly document cost function, constraints, Lagrangian.

    Inputs q and I are not standardized in this function,
    so they should be standardized beforehand 
    if standardized fitting is desired.
    At the provided constraint point, q_cons, 
    the returned polynomial will have slope dIdq_cons.
    """
    Ap = np.zeros( (order+1,order+1),dtype=float )
    b = np.zeros(order+1,dtype=float)
    # TODO: vectorize the construction of Ap
    for i in range(0,order):
        for j in range(0,order):
            Ap[i,j] = np.sum( q**j * q**i )
        Ap[i,order] = -1*i*q_cons**(i-1)
    for j in range(0,order):
        Ap[order,j] = j*q_cons**(j-1)
        b[j] = np.sum(I*q**j)
    b[order] = dIdq_cons
    p_fit = np.linalg.solve(Ap,b) 
    p_fit = p_fit[:-1]  # throw away Lagrange multiplier term 
    p_fit = p_fit[::-1] # reverse coefs to get np.polyfit format
    #from matplotlib import pyplot as plt
    #plt.figure(3)
    #plt.plot(q,I)
    #plt.plot(q,np.polyval(p_fit,q))
    #plt.plot(np.arange(q_cons,q[-1],q[-1]/100),np.polyval(p_fit,np.arange(q_cons,q[-1],q[-1]/100)))
    #plt.plot(q_cons,np.polyval(p_fit,q_cons),'ro')
    #plt.show()
    return p_fit
 
def spherical_normal_heuristics_setup():
    sigma_over_r = []
    width_metric = []
    intensity_metric = []
    qr0_focus = []
    # TODO: generate heuristics on a 1-d grid of sigma/r
    # instead of the 2-d grid being used here now.
    r0_vals = np.arange(10,41,10,dtype=float)              #Angstrom
    for ir,r0 in zip(range(len(r0_vals)),r0_vals):
        q = np.arange(0.001/r0,float(10)/r0,0.001/r0)       #1/Angstrom
        sigma_r_vals = np.arange(0*r0,0.21*r0,0.01*r0)      #Angstrom
        for isig,sigma_r in zip(range(len(sigma_r_vals)),sigma_r_vals):
            I = compute_spherical_normal_saxs(q,r0,sigma_r/r0) 
            I_at_0 = compute_spherical_normal_saxs(0,r0,sigma_r/r0) 
            d = saxs_spherical_normal_heuristics(q,I)
            sigma_over_r.append(float(sigma_r)/r0)
            qr0_focus.append(d['q_at_Iqqqq_min1']*r0)
            width_metric.append(d['pI_qwidth']/d['q_at_Iqqqq_min1'])
            intensity_metric.append(d['I_at_Iqqqq_min1']/I_at_0)
    # TODO: standardize before fitting, then revert after
    p_qr0_focus = np.polyfit(sigma_over_r,qr0_focus,2,None,False,None,False)
    p_w = np.polyfit(sigma_over_r,width_metric,2,None,False,None,False)
    p_I = np.polyfit(sigma_over_r,intensity_metric,3,None,False,None,False)
    print('polynomial for qr0 focus (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_qr0_focus[0],p_qr0_focus[1],p_qr0_focus[2]))
    print('polynomial for width metric (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_w[0],p_w[1],p_w[2]))
    print('polynomial for intensity metric (with x=sigma_r/r0): {}x^3 + {}x^2 + {}x + {}'.format(p_I[0],p_I[1],p_I[2],p_I[3]))
    plot = True
    if plot: 
        from matplotlib import pyplot as plt
        plt.figure(1)
        plt.scatter(sigma_over_r,width_metric)
        plt.plot(sigma_over_r,np.polyval(p_w,sigma_over_r))
        plt.figure(2)
        plt.scatter(sigma_over_r,intensity_metric)
        plt.plot(sigma_over_r,np.polyval(p_I,sigma_over_r))
        plt.figure(3)
        plt.scatter(sigma_over_r,qr0_focus)
        plt.plot(sigma_over_r,np.polyval(p_qr0_focus,sigma_over_r))
        plt.figure(4)
        plt.scatter(width_metric,intensity_metric) 
        plt.show()

def local_maxima_detector(y):
    """ 
    Finds local maxima in ordered data y.

    :param y: 1d numpy float array
    :return maxima: 1d numpy bool array

    *maxima* is *True* at a local maximum, *False* otherwise.

    This function makes no attempt to reject spurious maxima of any sort.
    """
    length = y.size
    greater_than_follower = np.zeros(length, dtype=bool)
    greater_than_leader = np.zeros(length, dtype=bool)
    greater_than_follower[:-1] = np.greater(y[:-1], y[1:])
    greater_than_leader[1:] = np.greater(y[1:], y[:-1])
    maxima = np.logical_and(greater_than_follower, greater_than_leader)
    # End points
    maxima[0] = greater_than_follower[0]
    maxima[-1] = greater_than_leader[-1]
    return maxima

def local_minima_detector(y):
    """
    Finds local minima in ordered data *y*.

    :param y: 1d numpy float array
    :return minima: 1d numpy bool array

    *minima* is *True* at a local minimum, *False* otherwise.

    This function makes no attempt to reject spurious minima of any sort.
    """
    return local_maxima_detector(-y)




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import copy

from PySide import QtCore

from .QWfWorker import QWfWorker
from .QWorkflow import QWorkflow
from ..core import pawstools
from ..core.operations import optools
from ..core.workflow.Workflow import Workflow
from ..core.operations.Operation import Operation, Batch, Realtime


class QWfManager(QtCore.QObject):
    """
    A Qt Signal-slot manager for paws Workflows.
    Takes a reference to a WfManager in the constructor.
    The QWfManager works mostly by
    calling on the methods of the WfManager.
    """

    def __init__(self,wfman,app):
        super(QWfManager,self).__init__()
        self.wfman = wfman
        self.app = app 
        self._n_threads = QtCore.QThread.idealThreadCount()
        self._n_wf_threads = 1
        self._wf_threads = dict.fromkeys(range(self._n_threads)[:self._n_wf_threads]) 
        self.qworkflows = OrderedDict() 
        self.wf_running = OrderedDict()

    def add_wf(self,wfname):
        """
        Add a QWorkflow to self.qworkflows, with key specified by wfname.
        If wfname is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
        """
        wf = Workflow()
        self.wfman.workflows[wfname] = wf
        self.qworkflows[wfname] = QWorkflow(wf)
        self.wf_running[wfname] = False
        self.wf_updated.emit(wfname)
        #wf.exec_finished.connect( partial(self.finish_wf,wfname) )
        #wf.setParent(self)
        #self.workflows[wfname] = wf
        # for every new workflow, add a plugin 
        #self.workflows[wfname] 
        #wf_pgin = WorkflowPlugin()
        #wf_pgin.inputs['workflow'] = wf
        #wf_pgin.start()
        #self.qplugman.add_plugin(wfname,wf_pgin)

    def load_from_dict(self,wfname,opman,opdict):
        """
        Create a workflow with name wfname.
        If wfname is not unique, self.workflows[wfname] is overwritten.
        Input opdict specifies operation setup,
        where each item in opdict provides enough information
        to get an Operation from OpManager opman
        and set up its Operation.input_locators.
        """
        self.add_wf(wfname)
        for opname, op_setup in opdict.items():
            op = self.wfman.build_op_from_dict(op_setup,opman)
            if isinstance(op,Operation):
                self.qworkflows[wfname].set_item(opname,op)
            else:
                self.wfman.write_log('Failed to load {} for workflow {}.'.format(opname,wfname))
        self.wf_updated.emit(wfname)

    def n_wf(self):
        return len(self.qworkflows)

    # this signal should emit the name
    # of the workflow that has been updated 
    wf_updated = QtCore.Signal(str)

    # this signal should emit the name
    # of the workflow that finished.
    wfdone = QtCore.Signal(str)

    #@QtCore.Slot(str)
    #def runWorkflow(self,wfname):
    #    self.run_wf(wfname)

    #@QtCore.Slot(str)
    #def stopWorkflow(self,wfname):
    #    self.stop_wf(wfname)

    def stop_wf(self,wfname):
        self.wf_running[wfname] = False
        self.wfdone.emit(wfname)

    def get_op(self,wfname,opname):
        return self.wfman.workflows[wfname].get_data_from_uri(opname)

    def run_wf(self,wfname):
        """
        Serially execute the operations of QWfManager.qworkflows[wfname].
        Uses optools.execution_stack() to determine execution order,
        and executes operations away from the main gui thread. 
        """
        self.wf_running[wfname] = True
        stk,diag = optools.execution_stack(self.wfman.workflows[wfname])
        for lst in stk:
            first_op = self.get_op(wfname,lst[0])
            batch_flag = isinstance(first_op,Batch)
            rt_flag = isinstance(first_op,Realtime)
            if not any([batch_flag,rt_flag]):
                self.execute_serial(wfname,lst)
            elif batch_flag:
                self.execute_batch(wfname,lst[0],lst[1])
            elif rt_flag:
                self.execute_realtime(wfname,lst[0],lst[1])
        self.wait_for_threads()
        # TODO: wait only for the threads working on wfname
        self.stop_wf(wfname)

    def execute_serial(self,wfname,op_list):
        thd_idx = self.next_available_thread()
        self.wfman.write_log('[{}] workflow {} running {}'
        .format(thd_idx,wfname,op_list))
        #for op_tag in op_list: 
        ops = [self.get_op(wfname,opname) for opname in op_list]
        for op in ops:
            optools.load_inputs(op,self.wfman.workflows[wfname],self.wfman.plugman)
        op_dict = OrderedDict(zip(op_list,ops))
        # Copy op_dict so that it can be thread-mobile
        op_dict = copy.deepcopy(op_dict)
        # Make a new Worker, give None parent so that it can be thread-mobile
        wf_wkr = QWfWorker(op_dict,None)
        wf_wkr.opDone.connect(partial(self.updateOperation,wfname))
        #wf_wkr.work()
        #wf_wkr.deleteLater()
        wf_thread = QtCore.QThread(self)
        wf_wkr.wfDone.connect(wf_thread.quit)
        wf_wkr.moveToThread(wf_thread)
        self.register_thread(thd_idx,wf_thread)
        wf_thread.started.connect(wf_wkr.work)
        wf_thread.finished.connect( partial(self.finish_thread,thd_idx,wfname) )
        wf_thread.finished.connect(wf_thread.deleteLater)
        wf_thread.finished.connect(wf_wkr.deleteLater)
        try:
            wf_thread.start()
        except Exception as ex:
            wf_wkr.deleteLater()
            wf_thread.quit()
            wf_thread.deleteLater()
            raise ex
        # Putting a processEvents() here as a shot in the dark
        # to address an intermittent freezing problem...
        self.app.processEvents()
        self.wait_for_thread(thd_idx)
        self.wf_updated.emit(wfname)
        # and another one here.
        self.app.processEvents()

    def execute_batch(self,wfname,batch_op_tag,batch_stk):
        # TODO: run this under a timer, take measures to speed it up
        batch_op = self.wfman.workflows[wfname].get_data_from_uri(batch_op_tag) 
        optools.load_inputs(batch_op,self.wfman.workflows[wfname],self.wfman.plugman)
        batch_op.run()
        self.qworkflows[wfname].set_op(batch_op_tag,batch_op)
        n_batch = len(batch_op.input_list())
        for i in range(n_batch):
            if self.wf_running[wfname]:
                input_dict = batch_op.input_list()[i]
                for uri,val in input_dict.items():
                    #import pdb; pdb.set_trace()
                    self.qworkflows[wfname].set_op_input_at_uri(uri,val)
                self.wfman.write_log( 'BATCH EXECUTION {} / {}'.format(i+1,n_batch) )
                for batch_lst in batch_stk:
                    self.execute_serial(wfname,batch_lst)
                saved_items_dict = OrderedDict()
                for uri in batch_op.saved_items():
                    # TODO # BUG: there is the chance for infinite recursion here
                    # if the batch is asked to save an upstream item?
                    save_data = self.wfman.workflows[wfname].get_data_from_uri(uri)
                    save_dict = self.wfman.uri_to_embedded_dict(uri,save_data) 
                    saved_items_dict = self.wfman.update_embedded_dict(saved_items_dict,save_dict)
                batch_op.output_list()[i] = saved_items_dict
                #self.qworkflows[wfname].set_item(batch_op_tag,batch_op)
                # Rather than re-setting the entire batch_op,
                # try just updating the outputs subtree for this iteration.
                batch_output_uri = batch_op_tag+'.'+optools.outputs_tag+'.'+batch_op.batch_outputs_tag()+'.'+str(i)
                self.qworkflows[wfname].tree_update_at_uri(batch_output_uri,saved_items_dict)
            else:
                raise pawstools.WorkflowAborted('[{}] {} was signaled to stop.'
                .format(__name__,wfname)) 

    def execute_realtime(self,wfname,rt_op_tag,rt_stk):
        msg = 'REALTIME EXEUCTION NOT YET IMPLEMENTED IN {}'.format(__name__)
        raise NotImplementedError(msg)

    @QtCore.Slot(str,str,Operation)
    def updateOperation(self,wfname,tag,op):
        self.qworkflows[wfname].set_item(tag,op)
        # processEvents() to kickstart the main event loop 
        # after finishing the update.
        # Application may freeze under some conditions
        # if this processEvents() is left out.
        self.app.processEvents()

    def finish_thread(self,th_idx,wfname):
        #print 'finishing thread {}'.format(th_idx)
        self._wf_threads[th_idx] = None
        self.wf_updated.emit(wfname)
        # processEvents() to kickstart the main event loop 
        # after finishing the update.
        # Application may freeze under some conditions
        # if this processEvents() is left out.
        self.app.processEvents()

    def register_thread(self,thread_idx,thread):
        self._wf_threads[thread_idx] = thread

    def wf_threads(self):
        return self._wf_threads

    def wait_for_thread(self,th_idx):
        """Wait for the thread at self._wf_threads[th_idx] to be finished"""
        #print 'waiting for thread {}'.format(th_idx)
        # when waiting for a thread to execute something,
        # processEvents() to ensure main event loop is current.
        self.app.processEvents()
        done = False
        interval = 1
        wait_iter = 0
        total_wait = 0
        while not done:
            #if wait_iter > 0:
            #    print '{}... still waiting for thread {} for {}ms'.format(wait_iter,th_idx,total_wait)
            done = True
            if self._wf_threads[th_idx] is not None:
                if not self._wf_threads[th_idx].isFinished():
                    done = False
                if not done:
                    if interval <= float(total_wait)*0.01 and interval <= 100:
                        interval = interval * 10
                    self.loopwait(interval)
                    wait_iter += 1
                    total_wait += interval

    def next_available_thread(self):
        for idx,th in self._wf_threads.items():
            if not th:
                #print '[{}] found available thread {}'.format(__name__,idx)
                return idx
        # if none found, wait for first thread in self.wfman.wf_threads 
        self.wait_for_thread(0)
        #print '[{}] falling back on thread 0'.format(__name__)
        return 0

    def wait_for_threads(self):
        """Wait for all workflow execution threads to finish"""
        for idx,th in self._wf_threads.items():
            self.wait_for_thread(idx)

    def loopwait(self,interval):
        """
        Create an event loop to delay some time without busywaiting.
        Time interval is specified in milliseconds.
        """
        l = QtCore.QEventLoop()
        t = QtCore.QTimer()
        t.setSingleShot(True)
        t.timeout.connect(l.quit)
        t.start(interval)
        l.exec_()
        # processEvents() to continue the main event loop while waiting.
        self.app.processEvents()





from matplotlib import pyplot as plt

from ..Operation import Operation
from .. import optools

class SimplePlot(Operation):
    """Plot a 1d array against another 1d array."""

    def __init__(self):
        input_names = ['x', 'y']
        output_names = ['figure', 'axis']
        super(SimplePlot, self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array; independent variable'
        self.input_doc['y'] = '1d array; dependent variable'
        self.output_doc['figure'] = 'matplotlib.pyplot.Figure with a plot of y vs x'
        self.output_doc['axis'] = 'matplotlib.pyplot.Axes object for use in plot modification'
        self.input_src['x'] = optools.wf_input
        self.input_src['y'] = optools.wf_input
        self.categories = ['DISPLAY']

    def run(self):
        self.outputs['figure'], self.outputs['axis'] = self.simple_plot(self.inputs['x'], self.inputs['y'])

    def simple_plot(x, y):
        fig, ax = plt.subplots(1)
        ax.plot(x, y)
        return fig, ax





import os
from functools import partial

from PySide import QtCore, QtGui, QtUiTools

from ..core.operations import optools
from ..core import pawstools
from ..core.models.ListModel import ListModel

class InputLoader(object):
    """This class controls the input_loader.ui to select inputs from various sources."""

    def __init__(self,name,src,src_manager=None,parent=None):
        super(InputLoader,self).__init__()
        self.input_name = name
        self.src = src
        self.src_manager = src_manager 
        self.setup_ui()
        if parent:
            self.ui.setParent(parent,QtCore.Qt.Window)

    def setup_ui(self):
        if self.src == optools.text_input:
            ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/text_input_loader.ui")
        elif self.src in [optools.plugin_input,optools.fs_input]:
            ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/tree_input_loader.ui")
        elif self.src == optools.wf_input:
            ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/wf_input_loader.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        self.ui.setWindowTitle("input loader")
        self.ui.input_box.setTitle(self.input_name)
        if self.src in [optools.plugin_input,optools.fs_input]:
            self.ui.source_treeview.setModel(self.src_manager)
        elif self.src == optools.wf_input:
            lm = ListModel(self.src_manager.qworkflows.keys())
            self.ui.wf_selector.setModel(lm)
            self.ui.wf_selector.currentIndexChanged.connect( partial(self.set_wf) )
        if self.src == optools.fs_input:
            self.src_manager.setRootPath(QtCore.QDir.currentPath())
            idx = self.src_manager.index(QtCore.QDir.currentPath())
            while idx.isValid():
                self.ui.source_treeview.setExpanded(idx,True)
                idx = idx.parent()
            self.ui.source_treeview.setExpanded(self.src_manager.index(pawstools.sourcedir),True)
            self.ui.source_treeview.hideColumn(1)
            self.ui.source_treeview.hideColumn(3)
            self.ui.source_treeview.setColumnWidth(0,250)
        elif self.src in [optools.plugin_input,optools.wf_input]:
            self.ui.source_treeview.hideColumn(2)
            self.ui.source_treeview.setColumnWidth(0,250)
            self.tree_model().set_all_flagged('select',False)
        if self.src == optools.plugin_input:
            self.ui.source_treeview.setRootIndex(self.src_manager.root_index())
        elif self.src == optools.wf_input:
            self.set_wf() 
        if self.src == optools.text_input:
            self.ui.set_button.setText("&Set entries")
            self.ui.add_button.setText("&Add entries")
        elif self.src in [optools.wf_input,optools.plugin_input,optools.fs_input]:
            self.ui.set_button.setText("&Set selected")
            self.ui.add_button.setText("&Add selected")
        self.values_model = ListModel()
        self.ui.values_list.setModel(self.values_model)
        self.ui.finish_button.setText("&Finish")
        self.ui.clear_button.setText("&Clear all")
        self.ui.remove_button.setText("&Remove selected")
        self.ui.list_toggle.setText("Package as &list")
        self.ui.add_button.clicked.connect( partial(self.make_selection,True) )
        self.ui.set_button.clicked.connect( partial(self.make_selection,False) )
        self.ui.remove_button.clicked.connect(self.rm_selection)
        self.ui.clear_button.clicked.connect(self.clear_values)

    def tree_model(self):
        if self.src in [optools.plugin_input,optools.fs_input]:
            return self.src_manager
        elif self.src == optools.wf_input:
            wf_idx = self.ui.wf_selector.currentIndex()
            wfname = self.ui.wf_selector.model().list_data()[wf_idx]
            return self.src_manager.qworkflows[wfname]
        else:
            return None

    def set_wf(self,wf_idx=None):
        if wf_idx is None:
            wf_idx = self.ui.wf_selector.currentIndex() 
        wfname = self.ui.wf_selector.model().list_data()[wf_idx]
        self.ui.source_treeview.setModel(self.src_manager.qworkflows[wfname])
        self.ui.source_treeview.setRootIndex(self.src_manager.qworkflows[wfname].root_index())
        self.ui.source_treeview.hideColumn(2)
        self.ui.source_treeview.setColumnWidth(0,250)

    def get_values(self):
        #import pdb; pdb.set_trace()
        if self.src == optools.text_input:
            vals = self.ui.source_textedit.toPlainText().split(os.linesep)
        elif self.src in [optools.plugin_input,optools.wf_input]:
            # Get all items currently selected
            idxs = self.tree_model().get_flagged_idxs('select')
            if any(idxs) and all([idx.isValid() for idx in idxs]):
                vals = [str(self.tree_model().get_uri_of_index(idx)).strip() for idx in idxs]
            # If user did not use tree selection model, take currentIndex()
            else:
                idx = self.ui.source_treeview.currentIndex()
                vals = [str(self.tree_model().get_uri_of_index(idx)).strip()]
            # Reset all selections
            self.tree_model().set_all_flagged('select',False)
        elif self.src == optools.fs_input:
            idx = self.ui.source_treeview.currentIndex()
            vals = [str(self.ui.source_treeview.model().filePath(idx)).strip()]
        return vals

    def make_selection(self,append=True):
        vals = self.get_values()
        if append:
            self.add_values(vals)
        else:
            self.set_values(vals)

    def clear_values(self):
        nvals = self.ui.values_list.model().rowCount()
        self.ui.values_list.model().removeRows(0,nvals)
        self.check_list_toggle()

    def set_values(self,vals):
        self.clear_values()
        self.add_values(vals)    
        self.check_list_toggle()

    def add_values(self,vals):
        if vals is not None:
            for val in vals:
                self.ui.values_list.model().append_item(str(val)) 
        self.check_list_toggle()

    def rm_selection(self):
        idx = self.ui.values_list.currentIndex()
        if idx.isValid():
            self.ui.values_list.model().remove_item(idx.row())
        self.check_list_toggle()

    def check_list_toggle(self):
        # if we have more than one selection, set and lock self.ui.list_toggle
        nvals = self.ui.values_list.model().rowCount()
        if nvals > 1:
            self.set_list_toggle()
            self.lock_list_toggle()
        # if we have one or less value, unset and unlock self.ui.list_toggle
        elif nvals <= 1:
            self.unset_list_toggle()
            self.unlock_list_toggle()

    def unset_list_toggle(self):
        self.ui.list_toggle.setCheckState(QtCore.Qt.Unchecked)
        
    def set_list_toggle(self):
        self.ui.list_toggle.setCheckState(QtCore.Qt.Checked)
        
    def lock_list_toggle(self):
        self.ui.list_toggle.setEnabled(False)

    def unlock_list_toggle(self):
        self.ui.list_toggle.setEnabled(True)




from matplotlib import pyplot as plt

from ..Operation import Operation
from .. import optools

class MPLFigFromXYData(Operation):

    def __init__(self):
        input_names = ['x', 'y']
        output_names = ['figure']
        super(MPLFigFromXYData, self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array; independent variable'
        self.input_doc['y'] = '1d array; dependent variable'
        self.output_doc['figure'] = 'matplotlib.pyplot.Figure with a plot of y vs x'
        self.input_src['x'] = optools.wf_input
        self.input_src['y'] = optools.wf_input
        self.categories = ['DISPLAY']

    def run(self):
        fig = plt.figure()
        plt.plot(self.inputs['x'],self.inputs['y'])
        self.outputs['figure'] = fig






from __future__ import print_function
import os
import pkgutil
import importlib
from collections import OrderedDict

import yaml

from .. import pawstools

# TODO: Deprecate ops.cfg configuration file.
# By default, inspect entire environment, 
# inventory all Operations, import none of them.
# Add operation environment specification to .wfl files.

# check for an ops.cfg file
cfg_file = pawstools.sourcedir+'/core/operations/ops.cfg'
if os.path.exists(cfg_file):
    load_flags = pawstools.load_cfg(cfg_file)
else:
    load_flags = OrderedDict() 
    
def save_config():
    """
    Call save_config() before closing 
    to save the state of which ops are enabled/disabled.
    """
    # Order the load flags using load_keys...
    od_load_flags = OrderedDict()
    for k in load_keys:
        od_load_flags[k] = load_flags[k]
    pawstools.save_cfg(od_load_flags,cfg_file)

# Keep track of keys that get loaded in this run.
# These keys are used to remove load_flags automatically
# when Operations or categories are renamed or removed.
load_keys = []

def update_load_flags():
    for k in load_flags.keys():
        if not k in load_keys:
            load_flags.pop(k)

def load_ops_from_path(path_,pkg,cat_root=''):
    ops = []
    cats = []
    # pkgutil.iter_modules returns module_loader, module_name, ispkg forall modules in path
    mods = pkgutil.iter_modules(path_)
    mods = [mod for mod in mods if mod[1] not in ['__init__','Operation','OpManager','optools','DMZ']]
    for modloader, modname, ispkg in mods:
        if cat_root == '':
            mod_root = modname 
        else:
            mod_root = cat_root+'.'+modname
        load_keys.append(mod_root)
        if mod_root in load_flags.keys():
            load_mod = bool(load_flags[mod_root])
        else:
            # NOTE: This line determines whether or not 
            # newly arrived modules should be loaded by default.
            load_mod = False
        # NOTE: Leaving all modules load_flag = True for now...
        #
        load_mod = True
        #
        load_flags[mod_root] = load_mod
        # if it is a package, recurse
        if load_mod and ispkg:
            pkg_path = [path_[0]+'/'+modname]
            pkg_ops, pkg_cats = load_ops_from_path(pkg_path,pkg+'.'+modname,mod_root)
            pkg_ops = [op for op in pkg_ops if not op in ops]
            pkg_cats = [cat for cat in pkg_cats if not cat in cats]
            ops = ops + pkg_ops
            cats = cats + pkg_cats
        elif load_mod:
            mod = importlib.import_module('.'+modname,pkg)
            # Get the operation from the module:
            # assume Operation name is same as module name
            try:
                op = getattr(mod,modname)
                ops.append( (cat_root,op) )
                if not cat_root in cats:
                    cats.append(cat_root)
            except AttributeError as ex:
                pass
            #    msg = str('Failed to load Operation subclass {} '.format(modname)
            #    + 'from module of the same name. Error message: '+ ex.message
            #    + '\nTo load an Operation, '
            #    + 'ensure the Operation subclass '
            #    + 'has the same name as its .py module file')
            #    print(msg)
    return ops, cats

def disable_ops(disable_root):
    # get all keys that contain disable_root
    disable_keys = [k for k in load_flags.keys() if disable_root in k]
    for k in disable_keys:
        load_flags[k] = False 

cat_op_list, cat_list = load_ops_from_path(__path__,__name__)

update_load_flags()

#op = load_op_from_module(mod,cat_root)
#mod_ops, mod_cats = load_ops_from_module(mod,cat_root)
#mod_ops = [op for op in mod_ops if not op in ops]
#mod_cats = [cat for cat in mod_cats if not cat in cats]
#ops = ops + mod_ops
#cats = cats + mod_cats

#def load_op_from_module(mod,cat_root):
    # Get the operation from the module:
    # assume Operation name is same as module name
    #op = getattr(mod,mod)
    #ops = []
    #cats = []
    #for nm, itm in mod.__dict__.items():
    #    try:
    #        # is it a class?
    #        if isinstance(itm,type):
    #            # is it a non-abstract subclass of Operation?
    #            if issubclass(itm,Operation) and not nm in ['Operation','Realtime','Batch']:
    #                op = getattr(mod,nm)
    #                ops.append( (cat_root,op) )
    #                if not cat_root in cats:
    #                    cats.append(cat_root)
    #                    #load_flags[cat_root] = True
    #                    #load_keys.append(cat_root)
    #                #cat = cat_root
    #                parent_cats_done = False
    #                while not parent_cats_done:
    #                    if not cat.rfind('.') == -1:
    #                        parcat = cat[:cat.rfind('.')]
    #                        if not parcat in cats:
    #                            cats.append(parcat)
    #                            load_keys.append(parcat)
    #                        cat = parcat
    #                    else:
    #                        parent_cats_done = True
    #                load_flags[nm] = True
    #                load_keys.append(nm)
    #    except ImportError as ex:
    #        print '[{}] had trouble dealing with {}: {}'.format(__name__,name,item)
    #        print 'Error text: {}'.format(ex.message)
    #        pass 
    #return ops, cats






import os
from datetime import datetime as dt

import yaml

# TODO: Make scratch directory and other cfg'ables into a cfg file

p = os.path.abspath(__file__)
# p = (pawsroot)/paws/core/pawstools.py
d = os.path.dirname(p)
# d = (pawsroot)/paws/core/
d = os.path.dirname(d)
# d = (pawsroot)/paws/
sourcedir = str(d)
d = os.path.dirname(d)
# d = (pawsroot)
#rootdir = str(d)
scratchdir = os.path.join(os.path.expanduser('~'),'.paws_scratch')
if not os.path.exists(scratchdir):
    os.mkdir(scratchdir)

# Get the code version from the paws_config.py file, store as __version__
with open(os.path.join(sourcedir,'paws_config.py')) as f: 
    exec(f.read())
version=__version__

class LazyCodeError(Exception):
    def __init__(self,msg):
        super(LazyCodeError,self).__init__(self,msg)

class WorkflowAborted(Exception):
    def __init__(self,msg):
        super(WorkflowAborted,self).__init__(self,msg)

def dtstr():
    """Return date and time as a string"""
    return dt.strftime(dt.now(),'%Y %m %d, %H:%M:%S')

def timestr():
    """Return time as a string"""
    return dt.strftime(dt.now(),'%H:%M:%S')

def update_file(filename,d):
    """
    Save the items in dict d into filename,
    without removing members not included in d.
    """
    if os.path.exists(filename):
        f_old = open(filename,'r')
        d_old = yaml.load(f_old)
        f_old.close()
        d_old.update(d)
        d = d_old
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()

def save_cfg(cfg_data,cfg_file):
    cfg = open(cfg_file,'w')
    yaml.dump(cfg_data,cfg)
    cfg.close()

def load_cfg(cfg_file):
    cfg = open(cfg_file,'r')
    cfg_data = yaml.load(cfg)
    cfg.close()
    if not cfg_data:
        cfg_data = OrderedDict() 
    return cfg_data




from collections import OrderedDict

import numpy as np
from scipy import interp
from scipy.optimize import minimize

from ...Operation import Operation
from ... import optools
from . import saxstools

class SpectrumParameterization(Operation):
    """
    The algorithm for guessing parameters
    for the size distributions of spherical nanoparticles 
    was developed and originally contributed by Amanda Fournier.    

    This operation uses a SAXS spectrum (I(q) vs. q), 
    along with some profiling information,
    to guess a set of parameters
    for fitting the SAXS spectrum
    to theoretical scattering models. 
    TODO: document scattering models.

    A precursor is modeled by a dilute, monodisperse spherical form factor.
    A spherical nanoparticle population is modeled by
    a discrete sum over a probability distribution
    of dilute, monodisperse spherical form factors. 
    A crystalline arrangement is modeled by 
    a sum of pseudo-Voigt peaks.

    Outputs a return code and heuristic guesses for SAXS model parameters. 
    Also outputs the theoretical result for I(q) with the guessed parameters.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'features', 'fixed_params', 'fixed_param_values']
        output_names = ['return_code','features','q_I_guess']
        super(SpectrumParameterization, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['features'] = str('dict of outputs from spectrum profiler operation. \n'
        + 'This operation makes use of the following keys: \n'
        + 'precursor_flag: Boolean for whether or not to include '
        + 'monodisperse spherical precursor parameters '
        + '(2 parameters: precursor radius and intensity scaling factor). \n'
        + 'form_flag: Boolean for whether or not to include '
        + 'normally-distributed spherical form factor parameters '
        + '(3 parameters: mean radius, standard deviation of radius, intensity scaling factor). \n'
        + 'structure_flag: Boolean for whether or not to include '
        + 'a set of Pseudo-Voigt diffraction peak parameters '
        + '(3 parameters: principal peak location, peak width, intensity scaling factor). \n'
        + 'form_id: string indicating the form factor '
        + '(options: sphere). \n'
        + 'structure_id: string indicating the structure to use in setting the diffraction peaks '
        + '(options: currently none, eventually fcc, hcp, bcc, etc).')
        self.input_doc['fixed_params'] = str('list of strings indicating which parameters '
        + 'should be fixed to specified values- these correspond one-to-one with fixed_param_values.')
        self.input_doc['fixed_param_values'] = 'list of values to which the corresponding fixed_params will be set.'

        self.output_doc['return_code'] = str('integer indicating whether or not '
        + 'the spectrum was successfully parameterized for the given flags. '
        + 'Possible values: 0=success, 1=error (raised an Exception), '
        + '2=warning (spectrum exhibits unexpected features).')
        self.output_doc['features'] = str('dict of features, same as input features, ' 
        + 'but updated with parameters for the flagged populations. ' 
        + 'Dict keys and descriptions: \n' 
        + 'I_at_0: Intensity at q=0, by fitting the low-q region to a polynomial with dI/dq(q=0)=0. \n' 
        + 'r0_pre: precursor term radius (Angstrom). \n'
        + 'I0_pre: precursor intensity scaling factor. \n'
        + 'r0_form: mean characteristic size (Angstrom) of form factor population. \n'
        + 'sigma_form: fractional standard deviation of characteristic size of form factor population \n'
        + 'I0_form: form factor scattering intensity scaling factor \n'
        + 'q0_structure: q value (1/Angstrom) for location of fundamental structure factor peak \n' 
        + 'I0_structure: structure factor intensity scaling factor \n'
        + 'sigma_structure: width parameter for Pseudo-Voigt diffraction peak profile') 
        self.output_doc['q_I_guess'] = str('n-by-2 array of q and the intensity spectrum '
        + 'corresponding to the population parameters in the features dict.')

        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['features'] = optools.wf_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type
        self.input_type['features'] = optools.ref_type

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1
        p = self.inputs['features'] 
        if p['bad_data_flag']:
            # sabotage
            self.outputs['return_code'] = 2
            p['PARAMETERIZE_ERROR'] = '[{}] found bad_data flag, aborting'.format(__name__)
            self.outputs['features'] = p
            return 
        pre_flag = p['precursor_flag']
        f_flag = p['form_flag']
        s_flag = p['structure_flag']
        fix_keys = self.inputs['fixed_params']
        fix_vals = self.inputs['fixed_param_values']
        fixed_params = {} 
        for k,v in zip(fix_keys,fix_vals):
            fixed_params[k] = v


        # Check for overconstrained system
        if 'I_at_0' in fix_keys and 'I0_form' in fix_keys and 'I0_pre' in fix_keys:
            val1 = fixed_params['I_at_0']
            val2 = fixed_params['I0_form'] + fixed_params['I0_pre'] 
            if not val1 == val2:
                msg = str('Spectrum intensity is overconstrained. '
                + 'I_at_0 is constrained to {}, '
                + 'but I0_form + I0_pre = {}. '.format(val1,val2))
                raise ValueError(msg)
        
        #if 'I_at_0' in fix_keys:
        #    if not fixed_params['I_at_0'] == 0:
        #        I_at_0 = fixed_params['I_at_0']
        #    else:
        #        # Perform a low-q spectrum fit to get I(q=0).
        #        # If there are form or structure factor terms,
        #        # use the lower 10% of the q range.
        #        if f_flag or s_flag:
        #            qmax_fit = q[0]+float(q[-1]-q[0])*0.1
        #            idx_lowq = (q<qmax_fit)
        #            I_at_0 = saxstools.fit_I0(q[idx_lowq],I[idx_lowq])
        #        else:
        #            I_at_0 = saxstools.fit_I0(q,I)
        if 'I0_form' in fix_keys and 'I0_pre' in fix_keys:
            I_at_0 = fixed_params['I0_pre'] + fixed_params['I0_form']
        else:
            # Perform a low-q spectrum fit to get I(q=0).
            # If there are form or structure factor terms,
            # use the lower 10% of the q range.
            if f_flag or s_flag:
                qmax_fit = q[0]+float(q[-1]-q[0])*0.1
                idx_lowq = (q<qmax_fit)
                I_at_0 = saxstools.fit_I0(q[idx_lowq],I[idx_lowq])
            else:
                I_at_0 = saxstools.fit_I0(q,I)
            if I_at_0 > 1.5*I[0] or I_at_0 < 0.5*I[0]:
                #sabotage 
                msg = 'polynomial fit for I(q=0) deviates too far from low-q values' 
                p['ERROR_MESSAGE'] = msg
                p['I_at_0'] = I_at_0
                p['bad_data_flag'] = True 
                self.outputs['features'] = p
                self.outputs['return_code'] = 1
                raise Exception(msg) 
        p['I_at_0'] = I_at_0

        r0_form = None
        sigma_form = None
        if f_flag:
            if ('r0_form' in fix_keys and 'sigma_form' in fix_keys):
                r0_form = fixed_params['r0_form']
                sigma_form = fixed_params['sigma_form']
            else:
                try:
                    r0_form, sigma_form = saxstools.spherical_normal_heuristics(q,I,I_at_0=I_at_0)
                except Exception as ex:
                    # sabotage
                    p['ERROR_MESSAGE'] = ex.message
                    p['bad_data_flag'] = True 
                    self.outputs['features'] = p
                    self.outputs['return_code'] = 1
                    raise ex 
                if 'r0_form' in fix_keys:
                    r0_form = fixed_params['r0_form']
                if 'sigma_form' in fix_keys:
                    sigma_form = fixed_params['sigma_form']
        p['r0_form'] = r0_form
        p['sigma_form'] = sigma_form
        
        r0_pre = None
        if pre_flag:
            if 'r0_pre' in fix_keys:
                r0_pre = fixed_params['r0_pre']
            else:
                r0_pre = saxstools.precursor_heuristics(q,I,I_at_0=I_at_0)
        p['r0_pre'] = r0_pre 

        I0_pre = None
        I0_form = None
        if pre_flag and f_flag:
            if 'I0_pre' in fix_keys and 'I0_form' in fix_keys:
                I0_pre = fixed_params['I0_pre']
                I0_form = fixed_params['I0_form']
            elif 'I0_pre' in fix_keys:
                I0_pre = fixed_params['I0_pre']
                I0_form = I_at_0 - I0_pre
            elif 'I0_form' in fix_keys:
                I0_form = fixed_params['I0_form']
                I0_pre = I_at_0 - I0_form
            else:
                I_pre = saxstools.compute_spherical_normal_saxs(q,r0_pre,0)
                I_form = saxstools.compute_spherical_normal_saxs(q,r0_form,sigma_form)
                I_nz = np.invert((I<=0))
                I_error = lambda x: np.sum( (np.log(I_at_0*(x*I_pre+(1.-x)*I_form)[I_nz])-np.log(I[I_nz]))**2 )
                x_res = minimize(I_error,[0.1],bounds=[(0.0,1.0)]) 
                x_fit = x_res.x[0]
                I0_form = (1.-x_fit)*I_at_0
                I0_pre = x_fit*I_at_0
        elif pre_flag:
            if 'I0_pre' in fix_keys:
                I0_pre = fixed_params['I0_pre']
            else:
                I0_pre = I_at_0 
        elif f_flag:
            if 'I0_form' in fix_keys:
                I0_form = fixed_params['I0_form'] 
            else:
                I0_form = I_at_0 
        p['I0_pre'] = I0_pre 
        p['I0_form'] = I0_form 

        q0_structure = None
        I0_structure = None
        sigma_structure = None
        #if s_flag:
        #   .....   
        p['q0_structure'] = q0_structure 
        p['I0_structure'] = I0_structure 
        p['sigma_structure'] = sigma_structure

        I_guess = saxstools.compute_saxs(q,p)
        q_I_guess = np.array([q,I_guess]).T
        self.outputs['q_I_guess'] = q_I_guess 
        self.outputs['features'] = p 
        self.outputs['return_code'] = 0

        #### Get logarithmic coef of determination to assess fit quality:
        #I_norm_nz = np.invert( (q_I_norm[:,1]<=0) )
        #logI_norm = np.log(q_I_norm[I_norm_nz,1])
        #logI_guess = np.log(q_I_guess[I_norm_nz,1])
        #sum_logvar = np.sum( (logI_norm-np.mean(logI_norm))**2 )
        #sum_logres = np.sum( (logI_guess-logI_norm)**2 ) 
        #R2_log = float(1)-float(sum_logres)/sum_logvar
        #p['R2_log'] = R2_log
        ####

        #from matplotlib import pyplot as plt
        #print self.outputs['features']
        #plt.figure(1)
        #plt.semilogy(q_I_norm[:,0],q_I_norm[:,1])
        #plt.semilogy(q_I_guess[:,0],q_I_guess[:,1],'r')
        #plt.show()







from collections import OrderedDict

from PySide import QtCore

from .QTreeSelectionModel import QTreeSelectionModel
from ..core.operations import optools

class QWorkflow(QTreeSelectionModel):
    """
    A QTreeSelectionModel for interacting with TreeModel Workflow.
    """

    def __init__(self,wf):
        flag_dict = OrderedDict()
        flag_dict['select'] = False
        flag_dict['enable'] = False
        super(QWorkflow,self).__init__(flag_dict,wf)
        #self._tree references wf after QTreeModel.__init__()
        #self.wf = wf

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} operation(s) loaded".format(self._tree._root_item.n_children())
        else:
            return super(QWorkflow,self).headerData(section,orientation,data_role)    

    wf_updated = QtCore.Signal()

    def remove_op(self,op_tag):
        """
        Remove an Operation from the workflow tree.
        """
        self.remove_item(op_tag)
        self.wf_updated.emit() 

    def set_op(self,op_tag,new_op):
        """
        Set workflow tree data at op_tag to new_op.
        """
        self.set_item(op_tag,new_op)
        self.wf_updated.emit() 

    def set_op_input_at_uri(self,uri,val):
        """
        Set an op input at uri to provided value val.
        The uri must be a valid uri in the QTreeModel,
        of the form opname.inputs.inpname.
        """
        path = uri.split('.')
        opname = path[0]
        if not path[1] == optools.inputs_tag:
            msg = '[{}] uri {} does not point to an input'.format(__name__,uri)
            raise ValueError(msg)
        inpname = path[2]
        #inp_uri = opname+'.'+optools.inputs_tag+'.'+inpname
        op = self._tree.get_data_from_uri(opname)
        op.input_locator[inpname].data = val
        self.set_item_at_uri(uri,val)

        # Additionally, set the op.input_locator[name].data to val.
        #opname = uri[:uri.find('.')]
        #inputname = uri[uri.rfind('.')+1:]
        #op = self._tree.get_data_from_uri(opname)
        #op.input_locator[inputname].data = val
        #self.wf_updated.emit() 

    #def update_op(self,op_tag,new_op):
    #    """Update given uri op_tag to Operation new_op."""
    #    self.set_item(op_tag,new_op)

    #    #p_idx = self.parent(rm_idx)
    #    #if not p_idx == self.root_index():
    #    #    msg = '[{}] Called remove_op on non-Operation at QModelIndex {}. \n'.format(__name__,rm_idx)
    #    #    raise ValueError(msg)
    #    #rm_itm = self.get_from_idx(rm_idx)
    #    #self.remove_item(rm_itm.tag,p_idx)


    #def tree_update(self,parent_itm,itm_tag,itm_data):
    #    if isinstance(itm_data,Operation):
    #        super(Workflow,self).tree_update(parent_itm,itm_tag,self.build_tree(itm_data))
    #    else:
    #        super(Workflow,self).tree_update(parent_itm,itm_tag,itm_data)






"""
Calibrate and reduce an image, given calibration parameters.

This module calls on PyFAI.AzimuthalIntegrator 
to calibrate an input image to I(q,chi),
and then reduce it to I(q).
"""

import numpy as np
import pyFAI

from ...Operation import Operation
from ... import optools

class CalReduce(Operation):
    """
    Input image data (ndarray) and a dict of .poni format calibration parameters
    Output q, I(q) 
    """
    def __init__(self):
        input_names = ['image_data','poni_dict']
        output_names = ['q','I','q_I']
        super(CalReduce,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_src['image_data'] = optools.wf_input
        self.input_src['poni_dict'] = optools.wf_input
        self.input_type['image_data'] = optools.ref_type
        self.input_type['poni_dict'] = optools.ref_type
        self.output_doc['q'] = 'Scattering vector magnitude q in 1/Angstrom.'
        self.output_doc['I'] = 'Integrated intensity at q.'
        self.output_doc['q_I'] = 'q and I zipped together an a n-by-2 numpy array.'

    def run(self):
        img = self.inputs['image_data']
        pd = self.inputs['poni_dict']
        p = pyFAI.AzimuthalIntegrator()
        p.setPyFAI(**pd)
        fpolz = pd['fpolz']
        # use a mask to screen negative pixels
        # mask should be 1 for masked pixels, 0 for unmasked pixels
        msk = np.ones(img.shape)*(img <= 0)
        q, I_of_q = p.integrate1d(img, 1000, radial_range=(0.0005,1.0005), mask=msk, polarization_factor=fpolz, unit='q_A^-1')
        # save results to self.outputs
        self.outputs['q'] = q
        self.outputs['I'] = I_of_q
        self.outputs['q_I'] = np.array([q,I_of_q]).T




from functools import partial

from PySide import QtCore, QtGui, QtUiTools

from ..core.operations import optools
from ..core.operations.Operation import Operation 
from ..core import pawstools
from ..core.models.ListModel import ListModel
from .InputLoader import InputLoader
from . import uitools

class WfUiManager(QtCore.QObject):
    # TODO: remove calls to workflow instance methods.
    # Replace with calls to workflow manager.

    def __init__(self,qwfman,qopman,qplugman):
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/wf_editor.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        # set self.ui to be deleted and to signal destroyed()
        # when its window is closed
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.qwfman = qwfman
        self.qopman = qopman 
        self.qplugman = qplugman 
        self.op = None
        self.setup_ui()
        # Dicts to keep track of input widgets, keyed by input variable names
        self.src_widgets = {} 
        self.type_widgets = {} 
        self.val_widgets = {} 
        self.btn_widgets = {} 
        self.input_loaders = {} 
        # Column definitions for the io layout        
        self.name_col = 0
        self.eq_col = 1
        self.src_col = 2
        self.type_col = 3
        self.val_col = 4
        self.btn_col = 5
        self.set_wf()

    def set_wf(self,wf_idx=None):
        if wf_idx is None:
            wf_idx = self.ui.wf_selector.currentIndex()
        wfname = self.ui.wf_selector.model().list_data()[wf_idx]
        self.ui.wf_browser.setModel(self.qwfman.qworkflows[wfname])
        self.ui.wf_browser.setRootIndex(self.qwfman.qworkflows[wfname].root_index())
        self.ui.wf_browser.hideColumn(1)
        self.ui.wf_browser.hideColumn(2)

    def current_wf(self):
        wfname = self.current_wfname()
        if wfname:
            return self.qwfman.qworkflows[wfname]

    def current_wfname(self):
        current_wf_idx = self.ui.wf_selector.currentIndex()
        if current_wf_idx > -1:
            return self.ui.wf_selector.model().list_data()[current_wf_idx]
        else:
            return None

    def get_op(self,src_qtree=None,itm_idx=QtCore.QModelIndex()):
        if src_qtree is None:
            src_qtree = self.current_wf()
        x = src_qtree.get_data_from_index(itm_idx)
        if x is not None:
            try:
                new_op_flag = issubclass(x,Operation)
            except:
                new_op_flag = False
            try:
                existing_op_flag = isinstance(x,Operation)
            except:
                existing_op_flag = False
            if not new_op_flag and not existing_op_flag:  
                self.op = None
                self.clear_io()
                #self.ui.op_info.setPlainText('Selected item: {}'.format(x))
            elif new_op_flag:
                # Create a new Operation 
                self.create_op(x)
            elif existing_op_flag:
                # Copy the setup of existing Operation
                op_dict = self.qwfman.wfman.op_setup_dict(x)
                # Get a new op with the same setup
                op_copy = self.qwfman.wfman.build_op_from_dict(op_dict,self.qopman.opman)
                op_tag = src_qtree.get_uri_of_index(itm_idx)
                self.set_op(op_copy,op_tag)

    def set_op(self,op,op_tag):
        """Set up ui elements around input op"""
        self.op = op
        #self.ui.op_info.setPlainText(self.op.description())
        self.build_io()
        self.ui.tag_entry.setText(op_tag)

    def create_op(self,op):
        """Instantiate op, call self.set_op()"""
        new_op = op()
        new_op_tag = self.current_wf()._tree.make_unique_uri(type(new_op).__name__)
        new_op.load_defaults()
        self.set_op(new_op,new_op_tag)

    def remove_op(self):
        """
        remove the selected operation from the workflow
        """
        idx = self.ui.wf_browser.currentIndex()
        if idx.isValid():
            itm = self.current_wf().get_from_index(idx) 
            while not itm.parent == self.current_wf()._tree._root_item:
                itm = itm.parent
            self.current_wf().remove_item(itm.tag)
            self.qwfman.wf_updated.emit(self.current_wfname())

    def load_op(self):
        """
        Package the finished Operation, put it in self.current_wf()
        """ 
        # Make sure all inputs are loaded
        for name in self.op.inputs.keys():
            self.set_input(name)
        tag = str(self.ui.tag_entry.text())
        if self.current_wf()._tree.is_tag_valid(tag):
            self.current_wf().set_item(tag,self.op)
            #import pdb; pdb.set_trace()
        else:
            # Request a different tag 
            msg_ui = uitools.message_ui(self.ui)
            msg_ui.setWindowTitle("Tag Error")
            msg_ui.message_box.setPlainText(self.current_wf()._tree.tag_error(tag))
            msg_ui.show()
        self.op = None
        self.clear_io()
        self.qwfman.wf_updated.emit(self.current_wfname())
        #replace_op_flag = self.current_wf().wf.contains_uri(tag)
        #if replace_op_flag: 
        #    self.current_wf().update_op(tag,self.op)
        #    self.op = None
        #else:
        #    self.current_wf().add_op(tag,self.op) 
        #    self.op = None
        #    self.clear_io()

    def set_input(self,name,src_ui=None):
        """
        Call build_input_locator to package an InputLocator for named input. 
        Store it in self.op.input_locator[name].
        This is called on all inputs when an Operation is loaded,
        so it should not alter an identical already-loaded input.
        """
        il = self.build_input_locator(name,src_ui)
        self.op.input_locator[name] = il
        # dereference the old input
        self.op.inputs[name] = None 

    def build_input_locator(self,name,ui=None):
        """
        Create an InputLocator for named input.
        If a ui is provided, it should be an input_loader.ui,
        and this should load data from that ui.
        """
        src = self.src_widgets[name].currentIndex()
        tp = self.type_widgets[name].currentIndex()
        il = None
        if src == optools.no_input:
            il = optools.InputLocator() 
        elif src == optools.batch_input:
            if tp == optools.auto_type:
                val = 'batch_input' 
            else:
                val = None
            il = optools.InputLocator(src,tp,val) 
        else:
            # all other input sources use a ui for loading
            if ui:
                val = self.fetch_from_input_ui(ui) 
                il = optools.InputLocator(src,tp,val)
                self.val_widgets[name].setText(str(il.val))
                ui.close()
                ui.deleteLater()
                # dereference the ui now that it is closed...
                self.input_loaders[name] = None
            else:
                # else, the input has already been loaded.
                # if src, tp, and val match, do nothing.
                val = self.val_widgets[name].text()
                if (self.op.input_locator[name].src == src 
                and self.op.input_locator[name].tp == tp 
                and str(self.op.input_locator[name].val) == val):
                    il = self.op.input_locator[name]
                else:
                    il = optools.InputLocator(src,tp,val)
        return il

    def fetch_from_input_ui(self,ui):
        # no matter the input source, 
        # ui.values_list.model().list_data()
        # should be a list of strings.
        val = ui.values_list.model().list_data()
        # if ui.list_toggle is not checked,
        # the value should be unpacked. 
        if not ui.list_toggle.isChecked():
            if len(val) == 0:
                val = None
            else:
                val = val[0]
        return val

    def fetch_data(self,name):
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        src = self.src_widgets[name].currentIndex()
        input_loader_title = self.ui.tag_entry.text()+'.'+optools.inputs_tag+'.'+name
        if src == optools.wf_input:
            inp_loader = InputLoader(input_loader_title,src,self.qwfman,self.ui)
            inp_loader.ui.wf_selector.setCurrentIndex(self.ui.wf_selector.currentIndex())
            inp_loader.set_wf()
        elif src == optools.fs_input:
            trmod = QtGui.QFileSystemModel()
            inp_loader = InputLoader(input_loader_title,src,trmod,self.ui)
        elif src == optools.plugin_input:
            inp_loader = InputLoader(input_loader_title,src,self.qplugman,self.ui)
        elif src == optools.text_input:
            inp_loader = InputLoader(input_loader_title,src,None,self.ui)
        if self.op.input_locator[name].src == src and self.op.input_locator[name].val is not None:
            if isinstance(self.op.input_locator[name].val,list):
                inp_loader.set_list_toggle()
                #for v in self.op.input_locator[name].val:
                inp_loader.add_values(self.op.input_locator[name].val)
            else:
                inp_loader.add_values([self.op.input_locator[name].val])
        inp_loader.ui.finish_button.clicked.connect( partial(self.set_input,name,inp_loader.ui) )
        self.input_loaders[name] = inp_loader
        #These do not need to be deleted on close. 
        #They will be deleted on clear_io() or when WfUiManager is closed.
        #inp_loader.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        inp_loader.ui.show()

    def clear_io(self):
        self.ui.op_name.setText('')
        self.ui.tag_entry.setText('')
        self.ui.load_button.setEnabled(False)
        # clean up internal objects via deleteLater()
        for nm,widg in (self.src_widgets.items()+self.type_widgets.items()
        +self.val_widgets.items()+self.btn_widgets.items()):
            if widg:
                widg.close()
                widg.deleteLater()
        for nm,il in self.input_loaders.items():
            if il:
                il.ui.close()
                il.ui.deleteLater()
        # dereference them to ensure they are collected
        self.src_widgets = {}
        self.type_widgets = {}
        self.val_widgets = {}
        self.btn_widgets = {}
        self.input_loaders = {}
        n_inp_widgets = self.ui.input_layout.count()
        for i in range(n_inp_widgets-1,-1,-1):
            item = self.ui.input_layout.takeAt(i)
            item.widget().close()
            item.widget().deleteLater()
        n_out_widgets = self.ui.output_layout.count()
        for i in range(n_out_widgets-1,-1,-1):
            item = self.ui.output_layout.takeAt(i)
            item.widget().close()
            item.widget().deleteLater()
    
    def reset_input_headers(self):
        if len(self.op.inputs) > 0:
            self.ui.input_layout.addWidget(uitools.r_hdr_widget('name'),0,self.name_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('source'),0,self.src_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('type'),0,self.type_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('value'),0,self.val_col,1,1)

    def reset_output_headers(self):
        if len(self.op.outputs) > 0:
            self.ui.output_layout.addWidget(uitools.r_hdr_widget('name'),0,self.name_col,1,1)
            self.ui.output_layout.addWidget(uitools.hdr_widget('description'),0,self.src_col,1,self.btn_col-self.src_col)

    def build_io(self):
        self.clear_io()
        self.reset_input_headers()
        self.reset_output_headers()
        self.ui.op_name.setText(type(self.op).__name__)
        self.ui.load_button.setEnabled(True)
        for i,name in zip(range(1,len(self.op.inputs)+1),self.op.inputs.keys()):
            # name 
            name_widget = uitools.name_widget(name)
            self.ui.input_layout.addWidget( name_widget,i,self.name_col,1,1 )
            # '=' 
            eq_widget = uitools.smalltext_widget('=')
            self.ui.input_layout.addWidget(eq_widget,i,self.eq_col,1,1)
            # source 
            src_widget = uitools.src_selection_widget()
            if self.op.input_locator[name] is not None:
                src = self.op.input_locator[name].src
            else:
                src = optools.no_input 
            src_widget.setCurrentIndex(src)
            self.ui.input_layout.addWidget( src_widget,i,self.src_col,1,1 )
            self.src_widgets[name] = src_widget 
            self.src_widgets[name].currentIndexChanged.connect( partial(self.reset_type_widget,name,i) )
            self.reset_type_widget(name,i,src)
        # Now handle outputs, much easier
        for i,name in zip(range(1,len(self.op.outputs)+1),self.op.outputs.keys()):
            # TODO: allow multiple lines for the desc_widgets.
            name_widget = uitools.name_widget(name)
            self.ui.output_layout.addWidget(name_widget,i,self.name_col)
            eq_widget = uitools.smalltext_widget('=')
            self.ui.output_layout.addWidget(eq_widget,i,self.eq_col)
            if self.op.output_doc[name]:
                desc_widget = uitools.bigtext_widget(self.op.output_doc[name])
            else:
                desc_widget = uitools.bigtext_widget('No description found.')
            self.ui.output_layout.addWidget(desc_widget,i,self.src_col,1,self.btn_col-self.src_col)

    def reset_type_widget(self,name,row,src=None):
        if name in self.type_widgets.keys():
            if self.type_widgets[name]:
                self.type_widgets[name].close()
        if not src:
            src = self.src_widgets[name].currentIndex()
        new_type_widget = uitools.type_selection_widget(src)
        if not self.op.input_locator[name].src == optools.no_input:
            if (self.op.input_locator[name].tp not in optools.invalid_types[src]
            and self.op.input_locator[name].src == src):
                # load operation defaults
                new_type_widget.setCurrentIndex(self.op.input_locator[name].tp)
            else:
                # set sensible defaults
                if src == optools.fs_input:
                    new_type_widget.setCurrentIndex(optools.path_type)
                elif src in [optools.wf_input,optools.plugin_input]:
                    new_type_widget.setCurrentIndex(optools.ref_type)
                elif src == optools.batch_input:
                    new_type_widget.setCurrentIndex(optools.auto_type)
        new_type_widget.currentIndexChanged.connect( partial(self.reset_val_widget,name,row,src) )            
        self.type_widgets[name] = new_type_widget  
        self.ui.input_layout.addWidget(new_type_widget,row,self.type_col,1,1)
        tp = self.type_widgets[name].currentIndex()
        self.reset_val_widget(name,row,src,tp)

    def reset_val_widget(self,name,row,src=None,tp=None):
        # TODO: Make val widgets expand height according to their content 
        if name in self.val_widgets.keys():
            if self.val_widgets[name]:
                self.val_widgets[name].close()
        if name in self.btn_widgets.keys():
            if self.btn_widgets[name]:
                self.btn_widgets[name].close()
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        if not src:
            src = self.src_widgets[name].currentIndex()
        if not tp:
            tp = self.type_widgets[name].currentIndex()
        btn_widget = QtGui.QPushButton()
        val_widget = QtGui.QLineEdit()
        #val_widget = QtGui.QTextEdit()
        if src == optools.no_input or tp == optools.none_type: 
            btn_widget.setText('no input')
            btn_widget.setEnabled(False)
            val_widget.setText('None')
        elif src == optools.batch_input:
            btn_widget.setText('auto')
            btn_widget.setEnabled(False)
            val_widget.setText('auto (batch)')
        elif src in [optools.wf_input,optools.fs_input,optools.plugin_input,optools.text_input]:
            btn_widget.setText('edit...')
            btn_widget.clicked.connect( partial(self.fetch_data,name) )
            if self.op.input_locator[name]:
                if self.op.input_locator[name].src == src:# and self.op.input_locator[name].tp == tp:
                    val_widget.setText(str(self.op.input_locator[name].val))
        val_widget.setReadOnly(True)
        self.ui.input_layout.addWidget(val_widget,row,self.val_col,1,1)
        self.ui.input_layout.addWidget(btn_widget,row,self.btn_col,1,1)
        self.val_widgets[name] = val_widget
        self.btn_widgets[name] = btn_widget

    def setup_ui(self):
        # LABELS and STYLES
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )
        self.ui.setWindowTitle("workflow setup")
        self.ui.input_box.setTitle("INPUTS")
        self.ui.output_box.setTitle("OUTPUTS")
        self.ui.finish_box.setTitle("FINISH / LOAD")
        self.ui.wf_box.setTitle("WORKFLOWS")
        self.ui.op_box.setTitle("OPERATIONS")
        self.ui.tag_prompt.setText('operation tag:')
        self.ui.tag_prompt.setReadOnly(True)
        self.ui.op_name.setText('-select an operation to begin setup-')
        self.ui.op_name.setAlignment(QtCore.Qt.AlignCenter)
        self.ui.tag_prompt.setAlignment(QtCore.Qt.AlignRight)
        self.ui.tag_prompt.setStyleSheet( "QLineEdit { background-color: transparent }" 
        + self.ui.tag_prompt.styleSheet() )
        self.ui.load_button.setText("&Finish")
        self.ui.rm_op_button.setText("&Remove selected operation")
        # SIGNALS, SLOTS, MODELS, VIEWS 
        lm = ListModel(self.qwfman.qworkflows.keys())
        self.ui.wf_selector.setModel(lm)
        self.ui.wf_selector.currentIndexChanged.connect(self.set_wf)
        #self.ui.wf_selector.currentIndexChanged.connect( partial(self.set_wf) )
        self.ui.wf_selector.activated.connect(self.set_wf)
        self.ui.wf_browser.clicked.connect( partial(self.get_op,None) )
        self.ui.rm_op_button.clicked.connect(self.remove_op)
        self.ui.op_selector.setModel(self.qopman)
        self.ui.op_selector.hideColumn(1)
        self.ui.op_selector.hideColumn(2)
        self.ui.op_selector.clicked.connect( partial(self.get_op,self.qopman) )
        self.ui.op_selector.clicked.connect( partial(uitools.toggle_expand,self.ui.op_selector) ) 
        self.ui.op_selector.setRootIndex(self.qopman.root_index())
        self.ui.load_button.clicked.connect(self.load_op)
        self.ui.load_button.setDefault(True)
        self.ui.load_button.setEnabled(False)
        # LAYOUT AND SIZING 
        szp = QtGui.QSizePolicy()
        szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        szp.setVerticalPolicy(QtGui.QSizePolicy.Preferred)
        szp.setHorizontalStretch(3)
        szp.setVerticalStretch(1)
        self.ui.input_frame.setSizePolicy(szp)
        finish_box_szp = QtGui.QSizePolicy()
        finish_box_szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        finish_box_szp.setVerticalPolicy(QtGui.QSizePolicy.Maximum)
        self.ui.finish_box.setSizePolicy(finish_box_szp)
        io_box_szp = QtGui.QSizePolicy()
        io_box_szp.setHorizontalPolicy(QtGui.QSizePolicy.Preferred)
        io_box_szp.setVerticalPolicy(QtGui.QSizePolicy.Maximum)
        self.ui.input_box.setSizePolicy(io_box_szp)
        self.ui.output_box.setSizePolicy(io_box_szp)
        self.ui.load_button.setMinimumWidth(100)
        self.ui.tag_prompt.setMaximumWidth(150)




import numpy as np

from ...Operation import Operation
from ...import optools

class ReadCSV_q_I_dI(Operation):
    """
    Read q, I, and (if available) dI from a csv-formatted file.
    If the csv has no third column, returns None for dI.
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['q','I', 'dI']
        super(ReadCSV_q_I_dI, self).__init__(input_names, output_names)
        self.input_doc['path'] = "path to .csv file"
        self.output_doc['q'] = "1d array, first column of csv, presumed to be scattering vector q"
        self.output_doc['I'] = "1d array, second column of csv, presumed to be scattering intensity I"
        self.output_doc['dI'] = "1d array, third column of csv, presumed to be error estimate of I"
        # source & type
        self.input_src['path'] = optools.fs_input
        self.input_type['path'] = optools.path_type

    def run(self):
        path = self.inputs['path']
        ncols = np.loadtxt(path, delimiter=',', skiprows=1).shape[1]
        if ncols == 2:
            q, I = np.loadtxt(path, dtype=float, delimiter=',', unpack=True, skiprows=1)
            dI = None
        elif ncols == 3:
            q, I, dI = np.loadtxt(path, dtype=float, delimiter=',', unpack=True, skiprows=1)
        #elif ncols == 4:
        #    col1, col2, col3, col4 = np.loadtxt(path, dtype=float, delimiter=',', unpack=True)
        else:
            raise ValueError("Input file has the wrong number of columns.  I don't know what to do with this.")
        self.outputs['q'] = q
        self.outputs['I'] = I
        self.outputs['dI'] = dI





"""
Operations config and processing routines
"""

import glob
import copy
from collections import Iterator
from collections import OrderedDict

# tags for inputs and outputs TreeItems    
inputs_tag = 'inputs'
outputs_tag = 'outputs'

# Declarations of valid sources and types for workflow and plugin inputs
##### TODO: the following but gracefully
no_input = 0
text_input = 1
fs_input = 2
wf_input = 3
plugin_input = 4
batch_input = 5 
valid_sources = [no_input,text_input,fs_input,wf_input,plugin_input,batch_input]
input_sources = ['none','text','filesystem','workflow','plugins','batch'] 

none_type = 0
str_type = 1
int_type = 2
float_type = 3
bool_type = 4
ref_type = 5
path_type = 6
auto_type = 7
valid_types = [none_type,str_type,int_type,float_type,bool_type,ref_type,path_type,auto_type]
input_types = ['none','string','integer','float','boolean','reference','path','auto']

invalid_types = {}                
invalid_types[no_input] = [str_type,int_type,float_type,bool_type,ref_type,path_type,auto_type]
invalid_types[text_input] = [ref_type,path_type,auto_type]
invalid_types[fs_input] = [str_type,int_type,float_type,bool_type,ref_type,auto_type]
invalid_types[wf_input] = [str_type,int_type,float_type,bool_type,auto_type]
invalid_types[plugin_input] = [str_type,int_type,float_type,bool_type,auto_type]
invalid_types[batch_input] = [str_type,int_type,float_type,bool_type,ref_type,path_type]

class InputLocator(object):
    """
    Objects of this class are used as containers for inputs to an Operation,
    and should by design contain the information needed to find the relevant input data.
    After the data is loaded, it should be stored in InputLocator.data.
    """
    def __init__(self,src=no_input,tp=none_type,val=None):
        self.src = src
        self.tp = tp
        self.val = val 
        self.data = None 

class FileSystemIterator(Iterator):

    def __init__(self,dirpath,regex):
        self.paths_done = []
        self.dirpath = dirpath
        self.rx = regex
        super(FileSystemIterator,self).__init__()

    def next(self):
        batch_list = glob.glob(self.dirpath+'/'+self.rx)
        for path in batch_list:
            if not path in self.paths_done:
                self.paths_done.append(path)
                return [path]
        return [None]

class ExecutionError(Exception):
    def __init__(self,msg):
        super(ExecutionError,self).__init__(self,msg)

def parameter_doc(name,value,doc):
    if isinstance(value, InputLocator):
        src_str = input_sources[value.src]
        tp_str = input_types[value.tp]
        v_str = str(value.val)
        return "- name: {} \n- source: {} \n- type: {} \n- value: {} \n- doc: {}".format(name,src_str,tp_str,v_str,doc) 
    else:
        tp_str = type(value).__name__
        return "- name: {} \n- type: {} \n- doc: {}".format(name,tp_str,doc) 

def get_uri_from_dict(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not isinstance(itm,dict):
            msg = 'something in {} is not a dict'.format(uri)
            raise KeyError(msg)
        if not k in itm.keys():
            msg = 'did not find uri {} in dict'.format(uri)
            raise KeyError(msg)
        else:
            itm = itm[k]
    return itm

def dict_contains_uri(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not k in itm.keys():
            return False
        else:
            itm = itm[k]
    return True


####### functions for fetching inputs and loading them into operations #######

def cast_type_val(tp,val):
    """
    Perform type casting for operation inputs.
    This should be called only for source = text_input.
    """
    if tp == none_type:
        val = None 
    elif tp == int_type:
        val = int(val)
    elif tp == float_type:
        val = float(val)
    elif tp == str_type:
        val = str(val)
    elif tp == bool_type:
        val = bool(eval(str(val)))
    else:
        msg = 'type selection {}, should be one of {}'.format(tp,valid_types)
        raise ValueError(msg)
    return val

def locate_input(il,wf=None,plugin_manager=None):
    """
    Return the data pointed to by a given InputLocator object.
    A Workflow and a PluginManager can be provided 
    as optional arguments,
    in which case they are used to fetch data.
    """
    if il.src == no_input or il.tp == none_type:
        return None
    elif il.src == batch_input:
        # Expect this input to have been set by Workflow Manager.
        return il.data 
    elif il.src == text_input: 
        if isinstance(il.val,list):
            return [cast_type_val(il.tp,v) for v in il.val]
        else:
            return cast_type_val(il.tp,il.val)
    elif il.src == wf_input:
        if il.tp == ref_type:

            # Note, this will return whatever data is stored in the TreeItem at uri.
            # If il.val is the uri of an input that has not yet been loaded,
            # this means it will get the InputLocator that currently inhabits that uri.

            # Note, this problem has now been fixed by changing Workflow.build_dict()
            # to not substitute InputLocators for inputs that had not been loaded.

            if isinstance(il.val,list):
                return [wf.get_data_from_uri(v) for v in il.val]
            else:
                return wf.get_data_from_uri(il.val)
        elif il.tp == path_type: 
            if isinstance(il.val,list):
                return [str(v) for v in il.val]
            else:
                return str(il.val)
    elif il.src == plugin_input:
        if il.tp == ref_type:
            if isinstance(il.val,list):
                return [plugin_manager.get_data_from_uri(v) for v in il.val]
            elif il.val is not None:
                return plugin_manager.get_data_from_uri(il.val)
            else:
                return None
        elif il.tp == path_type:
            if isinstance(il.val,list):
                return [str(v) for v in il.val]
            else:
                return str(il.val)
    elif il.src == fs_input:
        if isinstance(il.val,list):
            return [str(v) for v in il.val]
        else:
            return str(il.val)
    else: 
        msg = 'found input source {}, should be one of {}'.format(
        il.src, valid_sources)
        raise ValueError(msg)

def load_inputs(op,wf=None,plugin_manager=None):
    """
    Loads input data for an Operation from its input_locators.
    A Workflow and a PluginManager can be provided 
    as optional arguments,
    in which case they are used to fetch data.
    """
    for name,il in op.input_locator.items():
        if isinstance(il,InputLocator):
            il.data = locate_input(il,wf,plugin_manager)
            op.inputs[name] = il.data
        else:
            msg = '[{}] Found broken Operation.input_locator for {}: {}'.format(
            __name__, name, il)
            raise ValueError(msg)

####### functions having to do with workflow execution #######
# TODO: consider creating a wftools module instead

def get_valid_wf_inputs(op_tag,op):
    """
    Return the TreeModel uris of the op and its inputs/outputs 
    that are eligible as downstream inputs in the workflow.
    """
    # valid_wf_inputs should be the operation, its input and output dicts, and their respective entries
    valid_wf_inputs = [op_tag,op_tag+'.'+inputs_tag,op_tag+'.'+outputs_tag]
    valid_wf_inputs += [op_tag+'.'+outputs_tag+'.'+k for k in op.outputs.keys()]
    valid_wf_inputs += [op_tag+'.'+inputs_tag+'.'+k for k in op.inputs.keys()]
    return valid_wf_inputs
    
def stack_size(stk):
    sz = 0
    for lst in stk:
        for lst_itm in lst:
            if isinstance(lst_itm,list):
                sz += stack_size(lst_itm)
            else:
                sz += 1
    return sz

def stack_contains(itm,stk):
    for lst in stk:
        if itm in lst:
            return True
        for lst_itm in lst:
            if isinstance(lst_itm,list):
                if stack_contains(itm,lst_itm):
                    return True
    return False

def print_stack(stk):
    stktxt = ''
    opt_newline = '\n'
    for i,lst in zip(range(len(stk)),stk):
        if i == len(stk)-1:
            opt_newline = ''
        if len(lst) > 1:
            if isinstance(lst[1],list):
                substk = lst[1]
                stktxt += ('[\'{}\':\n{}\n]'+opt_newline).format(lst[0],print_stack(lst[1]))
            else:
                stktxt += ('{}'+opt_newline).format(lst)
        else:
            stktxt += ('{}'+opt_newline).format(lst)
    return stktxt

def is_op_ready(wf,op_tag,valid_wf_inputs,batch_routes=[]):
    op = wf.get_data_from_uri(op_tag)
    if op._batch_flag: 
        b_stk,op_rdy,diagnostics = batch_op_stack(wf,op_tag,valid_wf_inputs)
    elif op._realtime_flag: 
        rt_stk,op_rdy,diagnostics = batch_op_stack(wf,op_tag,valid_wf_inputs)
    else:
        inputs_rdy = []
        diagnostics = {} 
        for name,il in op.input_locator.items():
            msg = ''
            if (il.src == wf_input and il.tp == ref_type 
            and not il.val in valid_wf_inputs):
                inp_rdy = False
                msg = str('Operation input {}.inputs.{} (={}) '.format(op_tag,name,il.val)
                + 'not found in valid Workflow input list: {}'.format(valid_wf_inputs))
            elif (il.src == batch_input 
            and not op_tag+'.'+inputs_tag+'.'+name in batch_routes):
                inp_rdy = False
                msg = str('Operation input {}.inputs.{} (={}) '.format(op_tag,name,il.val)
                + 'expects batch input but is not listed in batch routes: {}'.format(batch_routes))
            else:
                inp_rdy = True
            inputs_rdy.append(inp_rdy)
            diagnostics[op_tag+'.'+inputs_tag+'.'+name] = msg
        if all(inputs_rdy):
            op_rdy = True
        else:
            op_rdy = False
    return op_rdy,diagnostics 

def batch_op_stack(wf,batch_op_tag,valid_wf_inputs):
    """
    Use batch_op.batch_ops() and a list of valid_wf_inputs 
    to build a stack (list) of lists of operations suitable for serial execution.
    """
    batch_op = wf.get_data_from_uri(batch_op_tag)
    # Batch and Realtime execution operations 
    # expect to have two inputs loaded by internal methods 
    # before calling batch_op.batch_ops() and batch_op.input_routes()
    batch_op.set_batch_ops(wf)
    batch_op.set_input_routes(wf)
    op_tags = batch_op.batch_ops()
    #load_inputs(batch_op,wf,plugin_manager)
    #op_tags = []
    #if batch_op._realtime_flag: 
    #    op_tags = batch_op.realtime_ops()
    #elif batch_op._batch_flag: 
    #    op_tags = batch_op.batch_ops()
    # make a copy of valid_wf_inputs
    # so that the existing valid_wf_inputs list is not mutated 
    valid_batch_inputs = copy.copy(valid_wf_inputs)
    # add the batch's own valid inputs to the list
    valid_batch_inputs += get_valid_wf_inputs(batch_op_tag,batch_op)
    # build the batch substack
    b_stk = []
    layer = []
    diagnostics = {}
    for op_tag in op_tags:
        op_rdy,op_diag = is_op_ready(wf,op_tag,valid_batch_inputs,batch_op.input_routes())
        diagnostics.update(op_diag)
        if op_rdy:
            layer.append(op_tag)
    while any(layer):
        b_stk.append(layer)
        for op_tag in layer:
            op = wf.get_data_from_uri(op_tag)
            valid_batch_inputs += get_valid_wf_inputs(op_tag,op)
        layer = []
        for op_tag in op_tags:
            op_rdy,op_diag = is_op_ready(wf,op_tag,valid_batch_inputs,batch_op.input_routes())
            diagnostics.update(op_diag)
            if op_rdy and not stack_contains(op_tag,b_stk):
                layer.append(op_tag)
    b_rdy = len(op_tags) == stack_size(b_stk) 
    return b_stk,b_rdy,diagnostics 

# TODO: the following
def check_wf(wf):
    """
    Check the dependencies of the workflow.
    Ensure that all loaded operations have inputs that make sense.
    Return a status code and message for each of the Operations.
    """
    pass

def execution_stack(wf):
    """
    Build a stack (list) of lists of Operation uris,
    such that each list indicates a set of Operations
    whose dependencies are satisfied by the Operations above them.
    For Batch or Realtime operations, 
    the layer should be of the form[batch_name,[batch_stack]],
    where batch_name indicates the batch controller Operation,
    and batch_stack is built from batch_op_stack().
    """
    stk = []
    valid_wf_inputs = []
    diagnostics = {}
    continue_flag = True
    while not stack_size(stk) == wf.n_ops() and continue_flag:
        ops_rdy = []
        ops_not_rdy = []
        for itm in wf._root_item.children:
            if not stack_contains(itm.tag,stk):
                op_rdy,op_diag = is_op_ready(wf,itm.tag,valid_wf_inputs)
                diagnostics.update(op_diag)
                if op_rdy:
                    ops_rdy.append(itm.tag)
                else:
                    ops_not_rdy.append(itm.tag)
        # Finished building list of ops currently ready. Now filter these into stack.
        if any(ops_rdy):
            # Which of these are not Batch/Realtime ops?
            non_batch_rdy = []
            for op_tag in ops_rdy:
                op = wf.get_data_from_uri(op_tag)
                if not any([op._batch_flag,op._realtime_flag]):
                    non_batch_rdy.append(op_tag)
            if any(non_batch_rdy):
                ops_rdy = non_batch_rdy
                stk.append(ops_rdy)
                for op_tag in ops_rdy:
                    op = wf.get_data_from_uri(op_tag)
                    valid_wf_inputs += get_valid_wf_inputs(op_tag,op)
            else:
                batch_tag = ops_rdy[0]
                ops_rdy = [batch_tag]
                batch_op = wf.get_data_from_uri(batch_tag)
                batch_stk,batch_rdy,batch_diag = batch_op_stack(
                wf,batch_tag,valid_wf_inputs)
                diagnostics.update(batch_diag)
                stk.append([batch_tag,batch_stk])
                valid_wf_inputs += get_valid_wf_inputs(batch_tag,batch_op)
        else:
            continue_flag = False
    return stk,diagnostics





from __future__ import print_function
import importlib
from collections import OrderedDict

from ..operations.Operation import Operation
from ..workflow.Workflow import Workflow
from ..operations import optools
from ..models.TreeModel import TreeModel
from .. import plugins as pgns
from .PawsPlugin import PawsPlugin

class PluginManager(TreeModel):
    """
    Tree structure for managing paws plugins.
    """

    def __init__(self,**kwargs):
        super(PluginManager,self).__init__()
        self.logmethod = print 

    def write_log(self,msg):
        self.logmethod(msg)
        #if self.logmethod:
        #else:
        #    print(msg)

    def plugin_setup_dict(self,pgin):
        pgin_mod = pgin.__module__[pgin.__module__.find('plugins'):]
        pgin_mod = pgin_mod[pgin_mod.find('.')+1:]
        dct = OrderedDict() 
        dct['plugin_module'] = pgin_mod
        dct[optools.inputs_tag] = pgin.inputs 
        return dct

    def load_from_dict(self,plugin_dict):
        """
        Load plugins from a dict that specifies their setup parameters.
        """
        for tag, pgin_spec in plugin_dict.items():
            pgin_uri = pgin_spec['plugin_module']
            pgin = self.get_plugin(pgin_uri)
            if pgin is not None:
                if not issubclass(pgin,PawsPlugin):
                    self.write_log('Did not find Plugin {} - skipping.'.format(pgin_uri))
                else:
                    pgin = pgin()
                    for name in pgin.inputs.keys():
                        if name in pgin_spec[optools.inputs_tag]:
                            pgin.inputs[name] = pgin_spec[optools.inputs_tag][name]
                    pgin.start()
                    # if already have this uri, first generate auto_tag
                    #if self.tree_contains_uri(uri):
                    #    uri = self.auto_tag(uri)
                    #self.add_plugin(uri,pgin)
                    self.set_item(tag,pgin)
            else:
                self.write_log('Did not find Plugin {} - skipping.'.format(ptype))

    #def tree_update(self,parent_idx,itm_tag,itm_data):
    #    if isinstance(itm_data,Operation) or isinstance(itm_data,PawsPlugin):
    #        super(PluginManager,self).tree_update(parent_idx,itm_tag,self.build_tree(itm_data))
    #    else:
    #        super(PluginManager,self).tree_update(parent_idx,itm_tag,itm_data)

    def build_tree(self,x):
        """
        Reimplemented TreeModel.build_tree() 
        so that TreeItems are built from PawsPlugins
        and Workflows and Operations.
        """
        if isinstance(x,PawsPlugin):
            #d = x.content()
            d = OrderedDict()
            for k,v in x.content().items():
                d[k] = self.build_tree(v)
        elif isinstance(x,Workflow):
            #d = x.op_dict()
            d = OrderedDict()
            for k,v in x.op_dict().items():
                d[k] = self.build_tree(v)
        elif isinstance(x,Operation):
            d = OrderedDict()
            d[optools.inputs_tag] = self.build_tree(x.inputs)
            d[optools.outputs_tag] = self.build_tree(x.outputs)
        else:
            return super(PluginManager,self).build_tree(x) 
        return d

    def get_plugin(self,pgin_type):    
        try:
            mod = importlib.import_module('.'+pgin_type,pgns.__name__)
            if pgin_type in mod.__dict__.keys():
                return mod.__dict__[pgin_type]
            else:
                msg = str('Did not find plugin {} in module {}'
                .format(pgin_type,mod.__name__))
                self.write_log(msg)
                return None 
        except Exception as ex:
            msg = str('Trouble loading module for plugin {}. '
            .format(pgin_name) + 'Error message: ' + ex.message)
            self.write_log(msg)
            return None

    def list_plugins(self):
        return self.root_tags()
        #r = self.get_from_idx(self.root_index())
        #return [itm.tag for itm in r.children]




import unittest

class TestAPI(unittest.TestCase):

    def test_start_stop(self):
        print 'testing paws api start/stop'
        import paws.api
        paw = paws.api.start()
        self.assertIsNone(paw._current_wf_name)

if __name__ == '__main__':
    unittest.main()




from os.path import splitext
from os import linesep
import numpy as np

from ...Operation import Operation
from ... import optools

class WriteCSV_q_I_dI(Operation):
    """Write q, I, and (if available) dI to a csv-formatted file."""

    def __init__(self):
        input_names = ['q','I','dI','image_location']
        output_names = ['csv_location']
        super(WriteCSV_q_I_dI, self).__init__(input_names, output_names)
        # docstrings
        self.input_doc['q'] = "1d ndarray; independent variable"
        self.input_doc['I'] = "1d ndarray; dependent variable; same shape as *q*"
        self.input_doc['dI'] = "1d ndarray; error estimate of *I*; same shape as *I*; if unavailable, use *None*"
        # source & type
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['dI'] = optools.wf_input
        self.input_src['image_location'] = optools.wf_input
        self.categories = ['OUTPUT.CSV']

    def run(self):
        csv_location = replace_extension(self.inputs['image_location'], '.csv')
        write_csv_q_I_maybe_dI(self.inputs['q'], self.inputs['I'], self.inputs['dI'], csv_location)
        self.outputs['csv_location'] = csv_location


def write_csv_q_I_maybe_dI(q, I, dI, nameloc):
    if dI is None:
        datablock = np.zeros((q.size,2),dtype=float)
        top_line = 'q, I'
    else:
        datablock = np.zeros((q.size,3),dtype=float)
        datablock[:,2] = dI
        top_line = 'q, I, dI'
    datablock[:,0] = q
    datablock[:,1] = I
    np.savetxt(nameloc, datablock, delimiter=',', newline=linesep, header=top_line)


def replace_extension(old_name, new_extension):
    '''
    Return a file name that is identical except for extension.

    :param old_name: string path or file name
    :param new_extension: string extension, e.g. ".txt"
    :return:

    Accepts extensions with or without an initial ".".
    '''
    # os.splitext gives the extension with '.' in front.
    # Rather than require the user to know this, I take care of it here.
    if new_extension[0] != '.':
        new_extension = '.' + new_extension
    root = splitext(old_name)[0]
    new_name = root + new_extension
    return new_name




from __future__ import print_function
from .. import operations as ops
from ..operations import optools
from ..models.TreeModel import TreeModel

class OpManager(TreeModel):
    """
    Tree structure for categorized storage and retrieval of Operations.
    """

    def __init__(self):
        super(OpManager,self).__init__()
        self.logmethod = print 
        self._n_ops = 0
        self._n_cats = 0

    def write_log(self,msg):
        self.logmethod(msg)

    def n_ops(self):
        return self._n_ops

    def load_cats(self,cat_list):
            for cat in cat_list:
                itm = self._root_item
                cat_tags = cat.split('.')
                cat_uri = cat_tags[0]
                if not self.contains_uri(cat_uri):
                    self.set_item(cat_uri,{})
                    self._n_cats += 1
                if len(cat_tags) > 1:
                    for cat_tag in cat_tags[1:]:
                        #if not cat_tag in self.list_child_tags(cat_uri):
                        cat_uri = cat_uri+'.'+cat_tag
                        if not self.contains_uri(cat_uri):
                            #print('set cat {}'.format(cat_uri))
                            self.set_item(cat_uri,{})
                            self._n_cats += 1

    def load_ops(self,cat_op_list):
        """
        Load OpManager tree from input cat_op_list.
        Format of cat_op_list is [(category1,op1),(category2,op2),...].
        i.e. each operation in cat_op_list is specified by a tuple,
        where the first element is a category,
        and the second element is the Operation itself.
        load_cats() should be called before load_ops()
        and should ensure that all cats in cat_op_list exist in the tree.
        """
        #### BUILD OPERATIONS TREE
        # Tree will consist of nodes indicating categories,
        # with subcategories or Operations as children.
        for cat_op in cat_op_list:
            self.add_op(cat_op[0],cat_op[1])
            self._n_ops += 1

    def add_op(self,cat,op):
        """
        Add op to the tree under category specified by uri cat.
        """
        self.set_item(cat+'.'+op.__name__,op)

    def remove_op(self,op_uri):
        """Remove op from the tree by its full category.opname uri"""
        self.remove_item(op_uri)
        self._n_ops -= 1



"""
Calibrate and reduce an image, given calibration parameters.

This module calls on PyFAI.AzimuthalIntegrator 
to calibrate an input image to I(q,chi).
"""

import numpy as np
import pyFAI

from ...Operation import Operation
from ... import optools


class Calibrate(Operation):
    """
    Input image data (ndarray) and a dict of calibration parameters 
    Return q, chi, I(q,chi) 
    """
    def __init__(self):
        input_names = ['image_data','poni_dict']
        output_names = ['q','chi','I_at_q_chi','I_at_q','q_I']
        super(Calibrate,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_src['image_data'] = optools.wf_input
        self.input_src['poni_dict'] = optools.wf_input
        self.input_type['image_data'] = optools.ref_type
        self.input_type['poni_dict'] = optools.ref_type
        self.output_doc['q'] = 'scattering vector magnitude in 1/Angstrom'
        self.output_doc['chi'] = 'azimuthal angle in degrees'     
        self.output_doc['I_at_q_chi'] = 'integrated intensity at q, chi'
        self.output_doc['I_at_q'] = 'chi-integrated intensity at q'
        self.output_doc['q_I'] = 'q and I_at_q zipped together into an n-by-2 numpy array'

    def run(self):
        img = self.inputs['image_data']
        pd = self.inputs['poni_dict']
        p = pyFAI.AzimuthalIntegrator()
        p.setPyFAI(**pd)
        fpolz = pd['fpolz']
        # use a mask to screen negative pixels
        # mask should be 1 for masked pixels, 0 for unmasked pixels
        s = int(img.shape[0])
        msk = np.ones((s,s))*(img <= 0)
        I_q_chi, q, chi = p.integrate2d(img, 1000, mask=msk, polarization_factor=fpolz, unit='q_A^-1')
        I_at_q = np.sum(I_q_chi,axis=0)
        #q = q * 1E9
        # save results to self.outputs
        self.outputs['q'] = q
        self.outputs['chi'] = chi 
        self.outputs['I_at_q_chi'] = I_q_chi 
        self.outputs['I_at_q'] = I_at_q 
        self.outputs['q_I'] = np.array([q,I_at_q]).T




import numpy as np

from scipy.optimize import curve_fit

from ...Operation import Operation
from ... import optools

class OptimizeSphericalDiffractionFit(Operation):
    """From an initial guess, optimize r0, I0, and fractional_variation."""

    # q, I, dI, qlim, guesses, log=False, clip=False, errors=False, baseline=False
    def __init__(self):
        input_names = ['q', 'I', 'dI', 'parameter_guesses','noise_floor','log_log_fit','exclude_high_q',
                       'error_weighting','q_upper_limit']
        output_names = ['fitted_parameters','I_fit','q_I_fit_logsafe','goodness_metrics']
        super(OptimizeSphericalDiffractionFit, self).__init__(input_names, output_names)
        # Documentation
        self.input_doc['q'] = '1d ndarray; wave vector values'
        self.input_doc['I'] = '1d ndarray; intensity values'
        self.input_doc['dI'] = '1d ndarray; intensity error estimate'
        self.input_doc['parameter_guesses'] = 'dictionary with keywords amplitude_at_zero, mean_size, and fractional_variation'
        self.input_doc['amplitude_at_zero'] = 'estimate of intensity at q=0'
        self.input_doc['mean_size'] = 'estimate of mean particle size'
        self.input_doc['fractional_variation'] = 'estimate of normal distribution sigma divided by mean size'
        self.input_doc['noise_floor'] = 'allow a fitted noise floor?'
        self.output_doc['fitted_parameters'] = 'dictionary with keywords amplitude_at_zero, mean_size, ' \
                                               'fractional_variation, and noise_magnitude'
        self.output_doc['I_fit'] = 'fitted intensity result'
        self.output_doc['q_I_fit_logsafe'] = 'fitted intensity result, zipped with q, logsafe ' \
                                             '(for display purposes; some points may be omitted)'
        self.output_doc['goodness_metrics'] = 'dictionary of various goodness-of-fit metrics'
        # Source and type
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['dI'] = optools.wf_input
        self.input_src['parameter_guesses'] = optools.wf_input
        self.input_src['noise_floor'] = optools.text_input
        self.input_src['log_log_fit'] = optools.text_input
        self.input_src['exclude_high_q'] = optools.text_input
        self.input_src['error_weighting'] = optools.text_input
        self.input_src['q_upper_limit'] = optools.text_input
        self.input_type['noise_floor'] = optools.bool_type
        self.input_type['log_log_fit'] = optools.bool_type
        self.input_type['exclude_high_q'] = optools.bool_type
        self.input_type['error_weighting'] = optools.bool_type
        self.input_type['q_upper_limit'] = optools.float_type
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type
        self.input_type['dI'] = optools.ref_type
        self.input_type['parameter_guesses'] = optools.ref_type
        # defaults
        self.inputs['noise_floor'] = False
        self.inputs['log_log_fit'] = True
        self.inputs['exclude_high_q'] = True
        self.inputs['error_weighting'] = False
        self.inputs['q_upper_limit'] = 0.25

    def run(self):
        q, I, dI = self.inputs['q'], self.inputs['I'], self.inputs['dI']
        guess_dict = self.inputs['parameter_guesses']
        log, clip, errors, baseline = self.inputs['log_log_fit'], self.inputs['exclude_high_q'], \
                                      self.inputs['error_weighting'], self.inputs['noise_floor']
        qlim = self.inputs['q_upper_limit']
        amplitude_at_zero, mean_size, fractional_variation = guess_dict['amplitude_at_zero'], guess_dict['mean_size'], \
                                                             guess_dict['fractional_variation']
        noise_magnitude = guess_noise_floor(q, I, mean_size)
        # computational validity checks
        valid = (~np.isnan(I)) & (~np.isinf(I)) & (I > 0)
        if (dI is not None) and errors:
            valid = valid & (~np.isnan(dI)) & (~np.isinf(dI)) & (dI > 0)
        if clip:
            valid = valid & (q <= qlim)
        x = q[valid]
        y = I[valid]
        if (dI is not None) and errors:
            dy = dI[valid]
        else:
            dy = None
        if log:
            y = safe_log(I[valid])
            if (dI is not None) and errors:
                # A reasonable proxy for log error
                dy = safe_log(I[valid] + dI[valid]) - safe_log(I[valid])
                # An alternative follows, prone to issues if (dI > I).any()
                # dI_val = 0.5 * (np.log(I_val + dI_val) - np.log(I_val - dI_val))
        # reasonable bounds; initial value list
        if baseline:
            bounds = ([amplitude_at_zero * 0.5, mean_size * 0.5, fractional_variation * 0.1, 0],
                      [amplitude_at_zero / 0.5, mean_size / 0.5, fractional_variation / 0.1, np.median(I)])
            init = [amplitude_at_zero, mean_size, fractional_variation, noise_magnitude]
        else:
            bounds = ([amplitude_at_zero * 0.5, mean_size * 0.5, fractional_variation * 0.1],
                      [amplitude_at_zero / 0.5, mean_size / 0.5, fractional_variation / 0.1])
            init = [amplitude_at_zero, mean_size, fractional_variation]
        # define function being fit to data
        if log:
            if baseline:
                def func(q, i0, r0, poly, noise):
                    return safe_log(generate_spherical_diffraction(q, i0, r0, poly) + noise)
            else:
                def func(q, i0, r0, poly):
                    return safe_log(generate_spherical_diffraction(q, i0, r0, poly))
        else:
            if baseline:
                def func(q, i0, r0, poly, noise):
                    return (generate_spherical_diffraction(q, i0, r0, poly) + noise)
            else:
                def func(q, i0, r0, poly):
                    return generate_spherical_diffraction(q, i0, r0, poly)
        # fit
        popt, pcov = curve_fit(func, x, y, init, dy, bounds=bounds)
        # unpack
        amplitude_at_zero, mean_size, fractional_variation = popt[0:3]
        if baseline:
            noise_magnitude = popt[3]
        else:
            noise_magnitude = None
        # feedback
        I_fit = amplitude_at_zero * blur(q * mean_size, fractional_variation)
        if noise_magnitude is not None:
            I_fit = I_fit + noise_magnitude
        chi_abs = chi_squared(I, I_fit)
        if dI is not None:
            chi_rel = chi_squared(I, I_fit, dI)
        else:
            chi_rel = None
        self.outputs['fitted_parameters'] = {'amplitude_at_zero':amplitude_at_zero, 'mean_size':mean_size,
                                             'fractional_variation':fractional_variation,
                                             'noise_magnitude':noise_magnitude}
        self.outputs['I_fit'] = I_fit
        self.outputs['q_I_fit_logsafe'] = logsafe_zip(q, I_fit)
        self.outputs['goodness_metrics'] = {'chi_absolute':chi_abs, 'chi_relative':chi_rel}


def safe_log(y):
    bads = (y <= 0) | (np.isnan(y))
    logy = np.log(y)
    logy[bads] = np.min(logy[~bads])
    return logy

def logsafe_zip(x, y):
    bad = (x <= 0) | (y <= 0) | np.isnan(y)
    x, y = x[~bad], y[~bad]
    x_y = np.zeros((x.size, 2))
    x_y[:, 0] = x
    x_y[:, 1] = y
    return x_y

def chi_squared(y1, y2, sigma=None):
    bads = (np.isnan(y1)) | (np.isnan(y2))
    if sigma is not None:
        bads = bads | (sigma == 0)
    else:
        sigma = np.ones(y1.shape)
    n = (~bads).sum()
    chi_2 = np.sum((y1[~bads] - y2[~bads])**2 * sigma[~bads]**-2) / (n - 1)
    return chi_2

def inclusive_arange(min, max, step):
    '''An inclusive-range version of np.arange.'''
    if np.mod((max - min), step) == 0:
        max += 0.5*step
    return np.arange(min, max, step)

def generate_spherical_diffraction(q, i0, r0, poly):
    x = q * r0
    i = i0 * blur(x, poly)
    return i

def generateRhoFactor(factor):
    '''
    Generate a guassian distribution of number densities (rho).

    :param factor: float
    :return:

    factor should be 0.1 for a sigma 10% of size
    '''
    factorCenter = 1
    factorMin = max(factorCenter-5*factor, 0.001)
    factorMax = factorCenter+5*factor
    factorVals = inclusive_arange(factorMin, factorMax, factor*0.02)
    # normalized gaussian:
    # ((sigma * (2 * np.pi)**0.5)**-1 )*np.exp(-0.5 * ((x - x0)/sigma)**2)
    rhoVals = ((factor * (2 * np.pi)**0.5)**-1 )*np.exp(-0.5 * ((factorVals - 1.)/factor)**2)
    return factorVals, rhoVals

def blur(x, factor):
    if factor == 0:
        return (3. * (np.sin(x) - x * np.cos(x)) * x**-3)**2
    factorVals, rhoVals = generateRhoFactor(factor)
    deltaFactor = factorVals[1] - factorVals[0]
    ysum = np.zeros(x.shape)
    for ii in range(len(factorVals)):
        effective_x = x / factorVals[ii]
        # spherical monodisperse diffraction:
        # (3. * (np.sin(x) - x * np.cos(x)) * x**-3)**2
        y = (3. * (np.sin(effective_x) - effective_x * np.cos(effective_x)) * effective_x**-3)**2
        ysum += rhoVals[ii]*y*deltaFactor
    return ysum

def guess_noise_floor(q, I, r0):
    qmin = q[0]
    qmax = q[-1]
    # we want q >> q1, ideally
    # the first dip is at 4.5 <~ x <~ 6, where x = q * r0.  We use pessimistic/strict option.
    qscale1 = 6/r0
    # and we know the worst signal to noise is at high q
    # so check the last tenth
    qscale2 = qmin + 0.9 * (qmax - qmin)
    if qscale2 < (qscale1 * 2): # just a sanity check, no real function
        print "Your data do not appear to be sampled to high q.  This is probably not a problem."
    selection = (q > qscale2)
    if selection.sum() < 10: # making sure there's a decent number of points in the sample
        print "Your data do not appear to be particularly well sampled.  This might be a problem."
        selection = np.zeros(q.size)
        selection[-10:] = True
    #noise = np.var(I[selection])
    noise = np.mean(I[selection])
    return noise




from os.path import splitext
from os import linesep
import numpy as np

from ...Operation import Operation
from ... import optools

class WriteArrayCSV(Operation):
    """Write a 2d array to a csv file"""

    def __init__(self):
        input_names = ['array','headers','dir_path','filename','filetag']
        output_names = ['csv_path']
        super(WriteArrayCSV, self).__init__(input_names, output_names)
        self.input_doc['array'] = 'any 2d array'
        self.input_doc['headers'] = 'list of headers (optional)- one header for each column of array'
        self.input_doc['dir_path'] = 'the path to the destination directory'
        self.input_doc['filename'] = 'the name of the file to be saved- no extension is expected'
        self.input_doc['filetag'] = 'tag appended to filename- no extension is expected'
        self.output_doc['csv_path'] = 'the path to the finished csv file: dir_path+filename+filetag+.csv'
        self.input_src['array'] = optools.wf_input
        self.input_src['headers'] = optools.text_input
        self.input_src['dir_path'] = optools.fs_input
        self.input_src['filename'] = optools.wf_input
        self.input_src['filetag'] = optools.text_input
        self.input_type['array'] = optools.ref_type
        self.input_type['headers'] = optools.str_type
        self.input_type['dir_path'] = optools.path_type
        self.input_type['filename'] = optools.ref_type
        self.input_type['filetag'] = optools.str_type
        self.inputs['filetag'] = ''

    def run(self):
        #import pdb; pdb.set_trace()
        h = self.inputs['headers']
        a = self.inputs['array']
        tag = ''
        if self.inputs['filetag']:
            tag = self.inputs['filetag']
        csv_path = self.inputs['dir_path']+'/'+self.inputs['filename']+tag+'.csv'
        self.outputs['csv_path'] = csv_path
        if h is not None:
            h_str = ''
            for i in range(len(h)-1):
                h_str += h[i] + ', '
            h_str = h_str+h[-1]
            np.savetxt(csv_path, a, delimiter=',', newline=linesep, header=h_str)
        else:
            np.savetxt(csv_path, a, delimiter=',', newline=linesep)




import numpy as np
import pypif.obj as pifobj

from ...Operation import Operation
from ... import optools

class PifNPSynth(Operation):
    """
    Package results from nanoparticle solution synthesis into a pypif.obj.ChemicalSystem object.
    """

    def __init__(self):
        input_names = ['uid_prefix','date_time','t_utc','temp_C','q_I','t_T','composition','features']
        output_names = ['pif']
        super(PifNPSynth,self).__init__(input_names,output_names)
        self.input_doc['uid_prefix'] = 'text string to prepend to pif uid (pif uid = uid_prefix+t_utc'
        self.input_doc['date_time'] = 'string date/time from measurement header file for pif record tags'
        self.input_doc['t_utc'] = 'time in seconds utc'
        self.input_doc['temp_C'] = 'temperature in degrees C'
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensities for saxs spectrum'
        self.input_doc['t_T'] = 'n-by-2 array of time (in seconds utc) and corresponding temperatures'
        self.input_doc['composition'] = 'dict describing recipe, in the format of IO.MISC.ReadNPSynthRecipe'
        self.input_doc['features'] = str('dict describing spectrum features, in the format of '
        + 'PROCESSING.SAXS.[SpectrumProfiler,SpectrumParameterization].')
        self.output_doc['pif'] = 'pif object representing the input data'
        self.input_src['uid_prefix'] = optools.text_input
        self.input_src['date_time'] = optools.wf_input
        self.input_src['t_utc'] = optools.wf_input
        self.input_src['temp_C'] = optools.wf_input
        self.input_src['q_I'] = optools.wf_input
        self.input_src['t_T'] = optools.wf_input
        self.input_src['composition'] = optools.wf_input
        self.input_src['features'] = optools.wf_input
        self.input_type['uid_prefix'] = optools.str_type
        self.input_type['date_time'] = optools.ref_type
        self.input_type['t_utc'] = optools.ref_type
        self.input_type['temp_C'] = optools.ref_type
        self.input_type['q_I'] = optools.ref_type
        self.input_type['t_T'] = optools.ref_type
        self.input_type['composition'] = optools.ref_type
        self.input_type['features'] = optools.ref_type

    def run(self):
        uid_pre = self.inputs['uid_prefix']
        t_str = self.inputs['date_time']
        t_utc = self.inputs['t_utc']
        uid_full = uid_pre+'_'+str(int(t_utc))
        temp_C = self.inputs['temp_C']
        q_I = self.inputs['q_I']
        t_T = self.inputs['t_T']
        c = self.inputs['composition']
        f = self.inputs['features']
        # Subsystems for solution ingredients
        #colloid_sys = pifobj.ChemicalSystem(uid_pre+'_pd_colloid',['colloidal Pd nanoparticles'],None,None,None,'Pd') 
        #acid_sys = pifobj.ChemicalSystem(uid_pre+'_oleic_acid',['oleic acid'],None,None,None,'C18H34O2') 
        #amine_sys = pifobj.ChemicalSystem(uid_pre+'_oleylamine',['oleylamine'],None,None,None,'C18H35NH2') 
        #TOP_sys = pifobj.ChemicalSystem(uid_pre+'_trioctylphosphine',['trioctylphosphine'],None,None,None,'P(C8H17)3')
        #subsys = [colloid_sys,acid_sys,amine_sys,TOP_sys]
        # TODO: Quantity information for subsystems
        main_sys = pifobj.ChemicalSystem()
        main_sys.uid = uid_full
        #main_sys.sub_systems = subsys
        if q_I is not None and temp_C is not None:
            main_sys.properties = self.saxs_to_pif_properties(q_I,temp_C)
        main_sys.tags = ['reaction id: '+uid_pre,'date: '+t_str,'utc: '+str(int(t_utc))]
        self.outputs['pif'] = main_sys

    def saxs_to_pif_properties(self,q_I,temp_C):
        #props = []
        #for i in range(len(q)):
        #pq = pifobj.Property()
        ### property: scattered intensity
        pI = pifobj.Property()
        pI.name = 'SAXS intensity'
        n_qpoints = q_I.shape[0]
        pI.scalars = [pifobj.Scalar(q_I[i,1]) for i in range(n_qpoints)]
        pI.units = 'counts'
        pI.conditions = []
        pI.conditions.append( pifobj.Value('SAXS scattering vector', 
                            [pifobj.Scalar(q_I[i,0]) for i in range(n_qpoints)],
                            None,None,None,'1/Angstrom') )
        pI.conditions.append(pifobj.Value('temperature',[pifobj.Scalar(temp_C)],None,None,None,'degrees Celsius'))
        return [pI] 
        
#    def make_piftemperature(self,t):
#        v = pifobj.Value()
#        v.name = 'temperature'
#        tscl = pifobj.Scalar()
#        tscl.value = str(t)
#        v.scalars = [tscl]
#        v.units = 'degrees Celsius'
#        return v
#
#    def make_pifvector(self,v):
#        pifv = []
#        for n in v:
#            s = pifobj.Scalar()
#            s.value = str(n)
#            pifv.append(s)
#        return pifv
#
#    def pifscalar(self,scl,errlo,errhi):
#        s = pifobj.Scalar()
#        s.value = str(scl) 
#        s.minimum = str(scl) 
#        s.maximum = str(scl) 
#        s.inclusiveMinimum = True
#        s.inclusiveMaximum = True
#        s.uncertainty = '+{},-{}'.format(errlo,errhi)
#        s.approximate = True
#        return s





import numpy as np

from ...Operation import Operation
from ...import optools

class CSVToXYData(Operation):
    """
    Read a csv-formatted file into x values and y values.
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['x','y','x_y']
        super(CSVToXYData, self).__init__(input_names, output_names)
        self.input_doc['path'] = 'path to .csv file'
        self.output_doc['x'] = 'numpy array built from first column of csv data'
        self.output_doc['y'] = 'numpy array built from second column of csv data'
        self.output_doc['x_y'] = 'n-by-2 numpy array with x and y values'
        # source & type
        self.input_src['path'] = optools.fs_input
        self.input_type['path'] = optools.path_type

    def run(self):
        path = self.inputs['path']
        x = np.loadtxt(path, dtype=float, delimiter=',', usecols=[0])
        y = np.loadtxt(path, dtype=float, delimiter=',', usecols=[1])
        self.outputs['x'] = x
        self.outputs['y'] = y
        x_y = np.array([x,y]).T
        self.outputs['x_y'] = x_y 
        





from ...Operation import Operation
from ... import optools
        
class CreateDataSet(Operation):
    """
    Take a Citrination client as input and use it to create a data set.
    Output the index of the created data set.
    """

    def __init__(self):
        input_names = ['client','name','description','share']
        output_names = ['ok_flag','dsid']
        super(CreateDataSet,self).__init__(input_names,output_names)
        self.input_doc['client'] = 'A reference to a running Citrination client.'
        self.input_doc['name'] = 'Name for the new data set.'
        self.input_doc['description'] = 'Description for the new data set.'
        self.input_doc['share'] = 'Flag whether or not dataset should be shared with all users on the instance.'
        self.output_doc['ok_flag'] = 'Indicator of whether or not the data set was created successfully.'
        self.output_doc['dsid'] = 'The index of the new data set, if created successfully.'
        self.input_src['client'] = optools.plugin_input
        self.input_src['name'] = optools.text_input
        self.input_src['description'] = optools.text_input
        self.input_src['share'] = optools.text_input
        self.input_type['client'] = optools.ref_type
        self.input_type['name'] = optools.str_type
        self.input_type['description'] = optools.str_type
        self.input_type['share'] = optools.bool_type
        self.inputs['description'] = 'New Citrination data set'
        self.inputs['share'] = False

    def run(self):
        c = self.inputs['client']
        f = True
        try:
            r = c.create_data_set()
            #if 'dataset' in r.keys():
            #    s = 'client successfully queried data set {}.'.format(dsid)
            #    f = True
            #else:
            #    s = 'client queried data set {} but found no dataset in response dict: Response: {}'.format(dsid,r)
            #    f = True
        except Exception as ex:
            s = 'client failed to create data set. Error message: {}'.format(ex.message)
            f = False
        self.outputs['ok_flag'] = f
        self.outputs['dsid'] = -1 




"""This package defines widgets that are used to communicate with paws."""

import numpy as np
from PySide import QtGui
from matplotlib.figure import Figure
from pypif.obj import System

from ...core.operations.Operation import Operation
from ...core.workflow.Workflow import Workflow 
from ...core.plugins.PawsPlugin import PawsPlugin
from .. import uitools
from .OpWidget import OpWidget
from .PifWidget import PifWidget
from .WorkflowGraphView import WorkflowGraphView
from .text_widgets import display_text, display_text_fast

if uitools.have_qt47:
    from .. import plotmaker_pqg as plotmaker
else:
    from .. import plotmaker_mpl as plotmaker

def make_widget(itm):
    if isinstance(itm,System):
        w = PifWidget(itm)
    elif isinstance(itm,Workflow):
        w = WorkflowGraphView(itm)
    elif isinstance(itm,PawsPlugin):
        w = QtGui.QTextEdit()
        msg = str('selected plugin is a {}. '.format(type(itm).__name__)
                + 'There are no display methods associated with this plugin. '
                + 'Add a display method to {} to view this plugin.'.format(__name__))
        w.setText(msg)
    elif isinstance(itm,Operation):
        w = OpWidget(itm)
    elif isinstance(itm,np.ndarray):
        dims = np.shape(itm)
        if len(dims) == 2 and dims[0] > 2 and dims[1] > 2:
            w = plotmaker.array_plot_2d(itm)
        elif len(dims) == 1 or (len(dims) == 2 and (dims[0]==2 or dims[1]==2)):
            w = plotmaker.array_plot_1d(itm)
    elif isinstance(itm,Figure):
        w = plotmaker.plot_mpl_fig(itm)
    elif (isinstance(itm,float) 
    or isinstance(itm,int) 
    or isinstance(itm,dict) 
    or isinstance(itm,list) 
    or isinstance(itm,str)):    
        t = display_text_fast(itm)
        w = QtGui.QTextEdit(t)
    else:
        msg = str('[{}]: selected item ({}) is not supported in {}'
            .format(__name__,type(itm).__name__,__name__)
            + '<br><br>Printout of item: <br>{}'.format(itm))
        w = QtGui.QTextEdit(msg)
    return w




